@@ -51,11 +51,12 @@
            ast:async-out-triggers
            ast:async?
            ast:async-port*
-           ast:call-statement
            ast:clr-events
            ast:component-model*
            ast:data*
            ast:declarative?
+           ast:default-value
+           ast:defer-variable*
            ast:direction
            ast:dotted-name
            ast:dzn-scope?
@@ -93,17 +94,21 @@
            ast:out-triggers
            ast:out-triggers-in-events
            ast:out-triggers-out-events
+           ast:provides-in-triggers
+           ast:provides-in-valued-triggers
+           ast:provides-in-void-triggers
+           ast:provides-out-triggers
            ast:provides-port
            ast:provides-port*
-           ast:provided-in-triggers
-           ast:provided-out-triggers
            ast:pure-funcq
            ast:rescope
            ast:req-events
-           ast:requires-port*
            ast:requires+async-port*
-           ast:required-in-triggers
-           ast:required-out-triggers
+           ast:requires-in-triggers
+           ast:requires-in-void-triggers
+           ast:requires-out-triggers
+           ast:requires-port*
+           ast:requires-no-injected-port*
            ast:rescope
            ast:scope
            ast:source-file
@@ -113,6 +118,7 @@
            ast:return-types-provides
            ast:return-values
            ast:valued-in-triggers
+           ast:values
            ast:void-in-triggers
            ast:out-triggers-valued-in-events
            ast:out-triggers-void-in-events
@@ -199,6 +205,7 @@
 
 (define-method (ast:argument* (o <action>)) ((compose ast:argument* .arguments) o))
 (define-method (ast:argument* (o <call>)) ((compose ast:argument* .arguments) o))
+(define-method (ast:argument* (o <defer>)) ((compose (cute and=> <> ast:argument*) .arguments) o))
 (define-method (ast:binding* (o <system>)) ((compose ast:binding* .bindings) o))
 (define-method (ast:event* (o <interface>)) ((compose ast:event* .events) o))
 (define-method (ast:event* (o <port>)) ((compose ast:event* .type) o))
@@ -223,6 +230,8 @@
 (define-method (ast:variable* (o <behavior>)) ((compose ast:variable* .variables) o))
 (define-method (ast:variable* (o <model>)) ((compose ast:variable* .behavior) o))
 (define-method (ast:variable* (o <compound>)) (filter (is? <variable>) (.elements o)))
+(define-method (ast:variable* (o <statement>)) '())
+(define-method (ast:variable* (o <variable>)) (list o))
 
 (define-method (ast:model* (o <model>))
   (define (ports o)
@@ -313,6 +322,9 @@
 (define-method (ast:provides? (o <trigger>))
   (and (.port.name o) ((compose ast:provides? .port) o)))
 
+(define-method (ast:provides? (o <action>))
+  (and (.port.name o) ((compose ast:provides? .port) o)))
+
 (define-method (ast:requires? (o <port>))
   (and (eq? (.direction o) 'requires) o))
 
@@ -369,6 +381,9 @@
 (define-method (ast:requires-port* (o <port>))
   (ast:requires-port* (parent o <component-model>)))
 
+(define-method (ast:requires-no-injected-port* (o <component-model>))
+  (filter (conjoin (negate ast:injected?) ast:requires?) (ast:port* o)))
+
 (define-method (ast:requires+async-port* (o <component-model>))
   (append (ast:requires-port* o) (ast:async-port* o)))
 
@@ -392,13 +407,21 @@
 (define-method (ast:out-event* o)
   (filter ast:out? (ast:event* o)))
 
-(define-method (ast:provided-in-triggers (o <component-model>))
+(define-method (ast:provides-in-triggers (o <component-model>))
   (map (cut trigger-in-component <> o)
        (append-map (lambda (port)
                      (map (lambda (event) (make <trigger> #:port.name (.name port) #:event.name (.name event) #:formals (ast:rescope ((compose .formals .signature) event) o)))
                           (filter ast:in? (ast:event* (.type port)))))
                    (filter ast:provides? (ast:port* o)))))
 
+(define-method (ast:provides-in-void-triggers (o <component-model>))
+  (filter (compose (is? <void>) .type .signature .event)
+          (ast:provides-in-triggers o)))
+
+(define-method (ast:provides-in-valued-triggers (o <component-model>))
+  (filter (compose (negate (is? <void>)) .type .signature .event)
+          (ast:provides-in-triggers o)))
+
 (define-method (ast:req-events (o <component>))
   (map (cut trigger-in-component <> o)
        (append-map (lambda (port)
@@ -415,7 +438,7 @@
                    (if (.behavior o) (ast:port* (.behavior o))
                        '()))))
 
-(define-method (ast:required-out-triggers (o <component-model>))
+(define-method (ast:requires-out-triggers (o <component-model>))
   (map (cut trigger-in-component <> o)
        (append-map (lambda (port)
                      (map (lambda (event) (make <trigger> #:port.name (.name port) #:event.name (.name event) #:formals (ast:rescope ((compose .formals .signature) event) o)))
@@ -437,7 +460,7 @@
                        '()))))
 
 (define-method (ast:in-triggers (o <component-model>))
-  (append (ast:provided-in-triggers o) (ast:required-out-triggers o) (ast:async-out-triggers o)))
+  (append (ast:provides-in-triggers o) (ast:requires-out-triggers o) (ast:async-out-triggers o)))
 
 (define-method (ast:in-triggers (o <interface>))
   (map (lambda (event) (make <trigger> #:event.name (.name event) #:formals (ast:rescope ((compose .formals .signature) event) o)))
@@ -491,22 +514,25 @@
 (define-method (ast:inout? (o <variable>))
   #f)
 
-(define-method (ast:provided-out-triggers (o <component-model>))
+(define-method (ast:provides-out-triggers (o <component-model>))
   (map (cut trigger-in-component <> o)
        (append-map (lambda (port)
                      (map (lambda (event) (make <trigger> #:port.name (.name port) #:event.name (.name event) #:formals (ast:rescope ((compose .formals .signature) event) o)))
                           (filter ast:out? (ast:event* (.type port)))))
                    (filter ast:provides? (ast:port* o)))))
 
-(define-method (ast:required-in-triggers (o <component-model>))
+(define-method (ast:requires-in-triggers (o <component-model>))
   (map (cut trigger-in-component <> o)
        (append-map (lambda (port)
                      (map (lambda (event) (make <trigger> #:port.name (.name port) #:event.name (.name event) #:formals (ast:rescope ((compose .formals .signature) event) o)))
                           (filter ast:in? (ast:event* (.type port)))))
                    (filter ast:requires? (ast:port* o) ))))
 
+(define-method (ast:requires-in-void-triggers (o <component-model>))
+  (filter (compose (is? <void>) ast:type) (ast:requires-in-triggers o)))
+
 (define-method (ast:out-triggers (o <component-model>))
-  (append (ast:provided-out-triggers o) (ast:required-in-triggers o)))
+  (append (ast:provides-out-triggers o) (ast:requires-in-triggers o)))
 
 (define-method (ast:void-in-triggers (o <component-model>))
   (filter
@@ -529,12 +555,12 @@
 
 (define-method (ast:out-triggers-void-in-events (o <component-model>))
   (filter
-   (lambda (t) (is-a? ((compose .type .signature .event) t) <void>))
+   (compose (is? <void>) .type .signature .event)
    (ast:out-triggers-in-events o)))
 
 (define-method (ast:out-triggers-valued-in-events (o <component-model>))
   (filter
-   (lambda (t) (not (is-a? ((compose .type .signature .event) t) <void>)))
+   (compose not (is? <void>) .type .signature .event)
    (ast:out-triggers-in-events o)))
 
 (define-method (ast:modeling? (o <event>))
@@ -643,7 +669,8 @@
        (equal? (.port.name a) (.port.name b))))
 
 (define-method (ast:equal? (a <field-test>) (b <field-test>))
-  (equal? (.field a) (.field b)))
+  (and (ast:equal? (.variable a) (.variable b))
+       (equal? (.field a) (.field b))))
 
 (define-method (ast:equal? (a <literal>) (b <literal>))
   (equal? (.value a) (.value b)))
@@ -657,6 +684,11 @@
    (ast:equal? (.left a) (.left b))
    (ast:equal? (.right a) (.right b))))
 
+(define-method (ast:equal? (a <unary>) (b <unary>))
+  (and
+   (eq? (class-of a) (class-of b))
+   (ast:equal? (.expression a) (.expression b))))
+
 (define-method (ast:equal? (a <expression>) (b <expression>))
   (if (eq? (class-of a) (class-of b))
       (throw 'add-ast:equal?-overload-for-type (class-of a))
@@ -754,16 +786,25 @@
          (index (list-index (cut ast:eq? o <>) formals)))
     index))
 
+(define-method (ast:defer-variable* (o <defer>))
+  (if (not (.arguments o)) (ast:member* (parent o <model>))
+      (map .variable (ast:argument* o))))
+
 (define-method (ast:expression->type (o <expression>))
-  (cond ((parent o <call>) ((compose .type ast:argument->formal) o))
-        ((is-a? o <action>) ((compose .type .signature .event) o))
-        ((parent o <assign>) => (compose .type .variable))
-        ((parent o <variable>) => .type)
-        ((parent o <return>) ((compose .type .signature) (parent o <function>)))
-        ((is-a? o <literal>) (ast:literal-value->type (.value o)))
-        ((is-a? o <bool-expr>) (make <bool>))
-        ((is-a? o <int-expr>) (make <int>))
-        (else (make <void>))))
+  (let ((p (.parent o)))
+    (define (as-p o class)
+      (or (as o class) (as p class)))
+    (cond ((as-p o <action>) => (compose .type .signature .event))
+          ((as-p o <assign>) => (compose .type .variable))
+          ((as-p o <arguments>) ((compose .type ast:argument->formal) o))
+          ((as-p o <var>) => (compose .type .variable))
+          ((as-p o <variable>) => .type)
+          ((as-p o <return>) => (compose .type .signature (cute parent <> <function>)))
+          ((is-a? o <bool-expr>) (make <bool>))
+          ((is-a? o <int-expr>) (make <int>))
+          ((parent (.parent o) <expression>) => ast:expression->type)
+          ((is-a? o <literal>) (ast:literal-value->type (.value o)))
+          (else (make <void>)))))
 
 (define-method (ast:return-types (interface <interface>))
   "Return all event types used in INTERFACE."
@@ -783,12 +824,34 @@
   "Return all event types used in COMPONENT."
   (delete-duplicates (append-map ast:return-types (filter-map ast:type (ast:provides-port* component))) ast:eq?))
 
+(define-method (ast:values (o <type>) void)
+  (cond
+   ((as o <void>)
+    void)
+   ((as o <enum>)
+    (let ((type (make <scope.name> #:ids (ast:full-name o))))
+     (map (cute make <enum-literal> #:type.name type #:field <>)
+          (ast:field* o))))
+   ((as o <bool>)
+    (map (cute make <literal> #:value <>) '("false" "true")))
+   ((as o <int>)
+    (map (cute make <literal> #:value <>)
+         (iota (1+ (- (.to (.range o)) (.from (.range o))))
+               (.from (.range o)))))))
+
+(define-method (ast:values (o <type>))
+  (ast:values o '()))
+
+(define-method (ast:default-value (o <type>))
+  (match (ast:values o)
+    ((default rest ...) default)))
+
+(define-method (ast:default-value (o <ast>))
+  (ast:default-value (ast:type o)))
+
 (define-method (ast:return-values (o <event>) void)
   (let ((type ((compose .type .signature) o)))
-    (cond ((as type <void>) void)
-          ((as type <enum>) (map (cut make <enum-literal> #:type.name (.name type) #:field <>) (ast:field* type)))
-          ((as type <bool>) (map (cut make <literal> #:value <>) '("true" "false")))
-          ((as type <int>) (map (cut make <literal> #:value <>) (iota (1+ (- (.to (.range type)) (.from (.range type)))) (.from (.range type))))))))
+    (ast:values type void)))
 
 (define-method (ast:return-values (o <event>))
   (ast:return-values o '()))
@@ -799,6 +862,9 @@
 (define-method (ast:return-values (o <action>))
   (ast:return-values (.event o)))
 
+(define-method (ast:return-values (o <trigger>))
+  (ast:return-values (.event o)))
+
 (define-method (ast:location (o <locationed>))
   (.location o))
 
@@ -1032,13 +1098,6 @@
     (or (and (null? statements) (not (is-a? (.parent o) <behavior>)))
         (and (pair? statements) ((compose ast:imperative? car) statements)))))
 
-(define (ast:call-statement o)
-  (match o
-    (($ <call>) o)
-    ((and ($ <assign>) (? (compose (is? <call>) .expression)) (= .expression call)) call)
-    ((and ($ <variable>) (? (compose (is? <call>) .expression)) (= .expression call)) call)
-    (_ #f)))
-
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; LOOKUP
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -1187,9 +1246,17 @@
 (define-method (ast:declaration* (o <behavior>))
   (append (ast:type* o) (ast:function* o) (ast:variable* o) (ast:port* o)))
 
+(define-method (ast:declaration* (o <defer>))
+  (tree-collect (is? <variable>) o))
+
 (define-method (ast:declaration* (o <compound>))
   (ast:variable* o))
 
+(define-method (ast:declaration* (o <if>))
+  (append (ast:variable* (.then o))
+          (or (and=> (.else o) ast:variable*) '())))
+
+
 (define-method (ast:declaration* (o <functions>))
   (ast:function* o))
 
@@ -1265,10 +1332,12 @@
       (and (is-a? (.parent o) <compound>)
            (ast:lookup-variable (.parent o) name (ast:statement-prefix o)))
       (ast:lookup-variable (.parent o) name statements)))
+    ((? (is? <defer>))
+     (ast:lookup-variable (.parent o) name (ast:statement-prefix o)))
     ((? (is? <if>))
      (or
       (and (is-a? (.parent o) <compound>)
-              (ast:lookup-variable (.parent o) name (ast:statement-prefix o)))
+           (ast:lookup-variable (.parent o) name (ast:statement-prefix o)))
       (ast:lookup-variable (.parent o) name statements)))
     (($ <function>)
      (or (find name? ((compose ast:formal* .signature) o))
@@ -1343,7 +1412,7 @@
 
 (define-method (.type (o <formal>))
   (let ((type-name (.type.name o)))
-    (if type-name (ast:lookup o (.type.name o))
+    (if type-name (ast:lookup (.parent (.parent o)) (.type.name o))
         (let ((formal (ast:event-formal o)))
           (and formal (.type formal))))))
 
@@ -1492,18 +1561,12 @@ to bottom."
     (or (not (null? (cdr scc)))
         (find (lambda (s) (eq? (.node o) (.node s))) (succ* o)))))
 
-
 (define-method (ast:function* (o <function>))
-  (let* ((compound (.statement o))
-         (statements (tree-collect ast:call-statement compound))
-         (functions (map (compose .function ast:call-statement) statements)))
-    (delete-duplicates (filter identity functions) ;; skip invalid calls
-                       (lambda (a b) (equal? (.name a) (.name b))))))
+  (let* ((calls (tree-collect (is? <call>) o))
+         (functions (filter-map .function calls)))
+    (delete-duplicates
+     functions
+     (lambda (a b) (equal? (.name a) (.name b))))))
 
 (define-method (ast:recursive? (o <function>))
-    (ast:graph-cyclic? ast:function* o))
-
-(define (ast-> ast)
-  ((compose
-    pretty-print
-    ) ast))
+  (ast:graph-cyclic? ast:function* o))
