@@ -36,6 +36,7 @@
 
   #:use-module ((oop goops) #:renamer (lambda (x) (if (member x '(<port> <foreign>)) (symbol-append 'goops: x) x)))
   #:use-module (dzn goops)
+  #:use-module (dzn display)
   #:use-module (dzn shell-util)
   #:use-module (dzn config)
 
@@ -43,8 +44,11 @@
   #:use-module (dzn code dzn)
   #:use-module (dzn normalize)
   #:use-module (dzn templates)
+  #:use-module (dzn vm goops)
 
-  #:export (%calling-context
+  #:export (<port-pair>
+           .other
+            %calling-context
             %queue-size
             %shell
             code
@@ -54,12 +58,16 @@
             code:annotate-shells
             code:arguments
             code:assign-reply
-            code:bind-provided
-            code:bind-required
+            code:bind-provides
+            code:bind-requires
+            code:capture-local
+            code:capture-member
             code:class-member?
             code:component-include
             code:component-port
             code:declarative-or-imperative
+            code:default-true
+            code:defer-condition
             code:enum-definer
             code:enum-field-definer
             code:enum-literal
@@ -84,10 +92,12 @@
             code:main-event-map-match-return
             code:main-out-arg
             code:main-out-arg-define
+            code:member-equality
             code:model
             code:non-injected-bindings
             code:non-injected-instances
             code:om
+            code:om+determinism
             code:ons
             code:out-argument
             code:port-bind?
@@ -99,7 +109,9 @@
             code:reply
             code:reply-type
             code:reply-types
+            code:requires-in-void-returns
             code:return
+            code:return-values
             code:trace-q-out
             code:trigger
             code:type-name
@@ -107,7 +119,9 @@
             code:used-foreigns
             code:variable->argument
             code:variable-name
-            string->enum-field))
+            string->enum-field)
+  #:re-export (.port
+               .port.name))
 
 ;; The calling-context to insert.
 (define %calling-context (make-parameter #f))
@@ -118,6 +132,16 @@
 ;; The name of the thread-safe shell.
 (define %shell (make-parameter #f))
 
+;;;
+;;; Ast extension.
+;;;
+(define-ast <port-pair> (<ast>)
+  (port)
+  (other))
+
+(define-method (.port.name (o <port-pair>)) (.name (.port o)))
+(define-method (.other.name (o <port-pair>)) (.name (.other o)))
+
 ;;;
 ;;; Top
 ;;;
@@ -130,28 +154,44 @@
          (models (map code:annotate-shells models)))
     models))
 
-(define-method (code:interface-include o)
-  (map (compose (cut make <file-name> #:name <>) code:file-name)
-       (delete-duplicates
-        (filter (compose (cut (negate equal?) (ast:source-file o) <>) ast:source-file .type)
-                (ast:port* o)))))
+(define-method (code:interface-include (o <top>) source-file)
+  (let* ((interfaces (ast:interface* o))
+         (files (filter (compose not
+                                 (cute equal? source-file <>)
+                                 ast:source-file)
+                        interfaces)))
+    (map (compose (cut make <file-name> #:name <>) code:file-name)
+         files)))
+
+(define-method (code:interface-include (o <top>))
+  (code:interface-include o (ast:source-file o)))
 
 (define-method (code:interface-include (o <foreign>))
-  (map (compose (cut make <file-name> #:name <>) code:file-name)
-       (filter (compose (cut (negate equal?) (ast:source-file (parent o <root>)) <>) ast:source-file)
-               (map .type (ast:port* o)))))
+  (code:interface-include o (ast:source-file (parent o <root>))))
 
 (define (code:component-include o)
- (filter (disjoin
-          (compose (is? <foreign>) .type)
-          (conjoin (compose ast:imported? .type) (lambda (i) (not (equal? (ast:source-file o)
-                                                                          (ast:source-file (.type i)))))))
-         (ast:instance* o)))
+  (let ((source-file (ast:source-file o)))
+    (filter (disjoin
+             (compose (is? <foreign>) .type)
+             (conjoin (compose ast:imported? .type)
+                      (compose not
+                               (cute equal? <> source-file)
+                               ast:source-file
+                               .type)))
+            (ast:instance* o))))
+
+(define-method (code:pump? (o <component>))
+  (and (or (pair? (ast:async-port* o))
+           (pair? (tree-collect (disjoin (is? <blocking>)
+                                         (is? <defer>)
+                                         (is? <blocking-compound>))
+                                o)))
+       o))
 
 (define-method (code:pump? (o <root>))
-  (filter (conjoin (negate ast:imported?) (is? <component>)
-                   (compose pair? ast:req-events))
-          (ast:model* o)))
+  (let ((components (filter (conjoin (negate ast:imported?) (is? <component>))
+                            (ast:model* o))))
+    (any code:pump? components)))
 
 
 ;;;
@@ -164,7 +204,7 @@
   (code:file-name (.type o)))
 
 (define-method (code:file-name (o <foreign>))
-  ((compose (cut string-join <> "_") ast:full-name) o))
+  (ast:full-name o))
 
 (define-method (code:file-name (o <ast>))
   (basename (ast:source-file o) ".dzn"))
@@ -253,9 +293,14 @@
       (make <local> #:name (.name o) #:type.name (.type.name o)
             #:expression (.expression o))))
 
+(define (make-out-formal formal)
+  (let* ((type (.type.name formal))
+         (out-formal (make <out-formal> #:name (.name formal) #:type.name type))
+         (out-formal (clone out-formal #:parent formal)))
+    (if (ast:in? formal) formal out-formal)))
+
 (define-method (code:variable-name (o <formal>))
-  (if (ast:in? o) o
-      (make <out-formal> #:name (.name o) #:type.name (.type.name o))))
+  (make-out-formal o))
 
 (define-method (code:variable-name (o <ast>))
   ((compose code:variable-name .variable) o))
@@ -340,6 +385,36 @@
     (if (ast:provides? port) port
         '())))
 
+(define-method (code:default-true (o <defer>))
+  (if (or (and=> (.arguments o) (compose null? .elements))
+          (null? (ast:variable* (parent o <component>)))) o
+          '()))
+
+(define-method (code:defer-condition (o <defer>))
+  (if (not (or (and=> (.arguments o)(compose null? .elements))
+               (null? (ast:variable* (parent o <component>))))) o
+      '()))
+
+(define-method (code:capture-local (o <defer>))
+  (let* ((references (tree-collect (disjoin(is? <assign>)
+                                           (is? <argument>)
+                                           (is? <var>))
+                                   (.statement o)))
+         (variables (map .variable references))
+         (local? (compose (cute ast:eq? <> o)
+                          (cute parent <> <defer>))))
+    (filter (negate (disjoin ast:member? local?))
+            variables)))
+
+(define-method (code:capture-member (o <defer>))
+  (ast:defer-variable* o))
+
+(define-method (code:capture-member (o <variable>))
+  o)
+
+(define-method (code:member-equality (o <defer>))
+  (filter (compose not (is? <extern>) .type) (ast:defer-variable* o)))
+
 
 ;;;
 ;;; Expressions
@@ -543,10 +618,6 @@
   (if (null? (code:injected-bindings o)) '()
       (list o)))
 
-;; (define-method (code:injected-instances (o <system>))
-;;   (if (null? (code:injected-bindings o)) '()
-;;       (injected-instances o)))
-
 (define (code:port-bind? bind)
   (and (code:port-binding? bind)
        bind))
@@ -572,7 +643,7 @@
 (define-method (code:non-injected-bindings (o <system>))
   (filter code:port-bind? (filter (negate injected-binding?) (ast:binding* o))))
 
-(define-method (code:bind-provided-required (o <binding>))
+(define-method (code:bind-provides-required (o <binding>))
   (let* ((model (parent o <model>))
          (left (.left o))
          (left-port (.port left))
@@ -582,15 +653,11 @@
         (cons left right)
         (cons right left))))
 
-(define-method (code:bind-provided (o <binding>))
-  ((compose car code:bind-provided-required) o))
+(define-method (code:bind-provides (o <binding>))
+  ((compose car code:bind-provides-required) o))
 
-(define-method (code:bind-required (o <binding>))
-  ((compose cdr code:bind-provided-required) o))
-
-(define-method (code:pump? (o <component>))
-  (if ((compose pair? ast:req-events) o) o
-      '()))
+(define-method (code:bind-requires (o <binding>))
+  ((compose cdr code:bind-provides-required) o))
 
 (define-method (code:trace-q-out o)
   (if ((compose ast:out? .event) o) o
@@ -602,18 +669,16 @@
 ;;;
 (define-method (code:main-out-arg (o <trigger>))
   (let* ((formals (ast:formal* o))
-         (formals (map
-                   (lambda (f) (if (ast:in? f) f (make <out-formal>)))
-                   formals)))
-    (map
-     (cute clone <> #:name <>)
-     formals (iota (length formals)))))
-
-(define-method (code:main-out-arg-define (o <trigger>))
-  (let ((formals (ast:formal* o)))
+         (formals (map make-out-formal formals)))
     (map (lambda (f i) (clone f #:name i))
          formals (iota (length formals)))))
 
+(define-method (code:main-out-arg-define (o <trigger>))
+  (let* ((formals (ast:formal* o))
+         (formals (map (lambda (f i) (clone f #:name i))
+                       formals (iota (length formals)))))
+    (filter (disjoin ast:out? ast:inout?) formals)))
+
 (define-method (code:main-out-arg-define-formal (o <formal>))
   (let ((type ((compose .value .type) o)))
     (if (ast:in? o) ""
@@ -660,13 +725,42 @@
              #:parent (.parent o))
       o))
 
+(define-method (code:return-values (o <component-model>))
+  (define (trigger->port-pairs trigger)
+    (map (cute make
+               <port-pair>
+               #:port (.port.name trigger)
+               #:other <>)
+         (map ->sexp (ast:return-values trigger))))
+  (let* ((triggers (filter (compose not (is? <void>) ast:type)
+                           (ast:requires-in-triggers o)))
+         (pairs (append-map trigger->port-pairs triggers)))
+    (delete-duplicates pairs (lambda (a b)
+                               (and (ast:eq? (.port a) (.port b))
+                                    (equal? (.other a) (.other b)))))))
+
+(define-method (code:requires-in-void-returns (o <component-model>))
+  (let ((triggers (ast:requires-in-void-triggers o)))
+    (delete-duplicates triggers
+                       (lambda (a b)
+                         (ast:eq? (.port a) (.port b))))))
+
 (define (code:om ast)
+  (let ((root ((compose
+                add-reply-port
+                normalize:event+illegals
+                remove-otherwise
+                (binding-into-blocking)
+                code:add-calling-context)
+               ast)))
+    (when (> (dzn:debugity) 1)
+      (ast:pretty-print root (current-error-port)))
+    root))
+
+(define (code:om+determinism ast)
   ((compose
-    add-reply-port
-    normalize:event
-    (remove-otherwise)
-    (binding-into-blocking)
-    code:add-calling-context)
+    add-determinism-temporaries
+    code:om)
    ast))
 
 
