@@ -41,6 +41,7 @@
 
   #:use-module ((oop goops) #:renamer (lambda (x) (if (member x '(<port> <foreign>)) (symbol-append 'goops: x) x)))
   #:use-module (dzn goops)
+  #:use-module (dzn display)
   #:use-module (dzn misc)
 
   #:use-module (dzn ast)
@@ -67,13 +68,6 @@
 (define %model-name (make-parameter #f))
 (define %next-alist (make-parameter #f))
 
-(define-ast <port-pair> (<ast>)
-  (port)
-  (other))
-
-(define-method (.port.name (o <port-pair>)) (.name (.port o)))
-(define-method (.other.name (o <port-pair>)) (.name (.other o)))
-
 
 ;;;
 ;;; Ticked root.
@@ -118,6 +112,9 @@
 ;;; Accessors.
 ;;;
 
+(define-method (makreel:behavior->defer-qout (o <behavior>))
+  (tree-collect (is? <defer>) o))
+
 (define-method (makreel:interface-reorder (o <behavior>))
   (if (parent o <interface>) o
       '()))
@@ -165,9 +162,10 @@
       (_ (mcrl2:process-continuation parent)))))
 
 (define (last-statement? o)
-  (let* ((p (parent o <scope>))
-         (statements (ast:statement* p)))
-    (ast:eq? (last statements) o)))
+  (or (not (is-a? (.parent o) <compound>))
+      (let* ((p (parent o <scope>))
+             (statements (ast:statement* p)))
+        (ast:eq? (last statements) o))))
 
 (define-method (makreel:assign-call-parameter (o <variable>))
   (if (last-statement? o) '()
@@ -175,8 +173,11 @@
 
 (define-method (makreel:assign-call-parameter (o <assign>))
   (if (and (last-statement? o)
-           (ast:eq? (parent (.variable o) <scope>) (parent o <scope>))) '()
-           o))
+           (not (let* ((cont (car (makreel:continuation o)))
+                       (variables (variables-in-scope cont)))
+                  (member (.variable o) variables ast:eq?))))
+      '()
+      o))
 
 ;; FIXME: non-compatible copy from mcrl2 scope vs model ticking:
 ;;  <scope-name (IConsole) State'> vs <interface IConsole'>
@@ -590,6 +591,9 @@
 (define-method (makreel:proc (o <function>))
   (makreel:proc-list o))
 
+(define-method (makreel:proc (o <defer>))
+  (makreel:proc-list (.statement o)))
+
 (define-method (makreel:proc-assign (o <assign>))
   (let ((expression (.expression o)))
     (if (or (is-a? expression <action>)
@@ -639,22 +643,28 @@
     (ast:variable* behavior)))
 
 (define-method (makreel:locals- (o <ast>))
-  (if ((is? <behavior>) o) '()
-      (let* ((parent (.parent o)))
-        (cond ((is-a? parent <compound>)
-               (let ((pre (cdr (member o (reverse (ast:statement* parent)) ast:eq?))))
-                 (append (filter (is? <variable>) pre) (makreel:locals parent))))
+  (if (is-a? o <behavior>) '()
+      (let* ((p (.parent o)))
+        (cond ((is-a? p <compound>)
+               (let ((pre (cdr (member o (reverse (ast:statement* p)) ast:eq?))))
+                 (append (filter (is? <variable>) pre) (makreel:locals p))))
+              ((is-a? p <defer>)
+               (makreel:locals- p))
               ((is-a? o <function>) ((compose ast:formal* .signature) o))
-              (else (makreel:locals parent))))))
+              (else (makreel:locals p))))))
 
 (define (makreel:locals o)
   (define (locals root o)
     (makreel:locals- o))
-  ((ast:pure-funcq locals) (parent o <root>) o))
+  (let* ((model (parent o <model>))
+         (root (parent model <root>)))
+    ((ast:pure-funcq locals) (list root model) o)))
 
 (define-method (variables-in-scope (o <model>)) (members o))
 (define-method (variables-in-scope (o <ast>))
-  (let ((stack (and (parent o <function>) (clone (make <stack>) #:parent o))))
+  (let ((stack (and (parent o <function>)
+                    (not (parent (.parent o) <defer>))
+                    (clone (make <stack>) #:parent o))))
     (append (members o) (makreel:locals o) (if stack (list stack) '()))))
 
 (define-method (makreel:stack-parameters (o <ast>))
@@ -716,11 +726,11 @@
   (define (statement-continuation o)
     (let* ((p (.parent o))
            (cont (cdr (member o (ast:statement* p) (lambda (a b) (eq? (.node a) (.node b)))))))
-      (if (pair? cont)
-          (list (car cont))
+      (if (pair? cont) (list (car cont))
           (let ((grandp (.parent p)))
             (match grandp
               (($ <compound>) (statement-continuation p))
+              (($ <defer>) (list (parent o <behavior>)))
               ((? (is? <declarative>)) (list (parent o <behavior>)))
               (_  (makreel:continuation grandp)))))))
 
@@ -921,23 +931,25 @@
   (map .expression (filter (compose (is? <int>) .type) (ast:member* o))))
 
 (define-method (makreel:stack? (o <call>))
-  (or (and (parent o <function>) o) '()))
+  (or (and (not (parent o <defer>)) (parent o <function>) o) '()))
 
 (define-method (makreel:stack-empty? (o <call>))
-  (or (and (not (parent o <function>)) o) '()))
+  (or (and (or (parent o <defer>) (not (parent o <function>))) o) '()))
 
 (define-method (makreel:stack-destructor (o <ast>))
-  (let ((f (parent o <function>))
+  (let ((f (and (not (parent o <defer>))
+                (parent o <function>)))
         (locals (makreel:locals o))
         (return-value (if (and (or (is-a? o <assign>) (is-a? o <variable>))
                                (is-a? (.expression o) <call>)) (list (ast:type o))
                           '())))
-    (if f (append locals (list f) return-value)
+    (if f (append locals (list f)
+                  return-value)
         (append locals return-value))))
 
 (define-method (makreel:process-argument-stack? (o <call>))
   (let ((function (parent o <function>)))
-    (cond ((and function (.last? o)) #f)
+    (cond ((and function (.last? o)) function)
           (else o))))
 
 (define-method (makreel:enum-name (o <enum>))
@@ -951,15 +963,19 @@
 (include-from-path "dzn/templates/makreel.scm")
 
 (define (makreel:om ast)
-  ((compose
-    makreel:mark-tail-call
-    add-function-return
-    normalize:state
-    (remove-otherwise)
-    makreel:tick-names
-    add-explicit-temporaries
-    purge-data
-    ) ast))
+  (let ((root ((compose
+                makreel:mark-tail-call
+                add-function-return
+                normalize:state+illegals
+                remove-otherwise
+                makreel:tick-names
+                add-explicit-temporaries
+                add-defer-end
+                purge-data
+                ) ast)))
+    (when (> (dzn:debugity) 1)
+      (ast:pretty-print root (current-error-port)))
+    root))
 
 
 ;;;
