@@ -2,7 +2,7 @@
 ;;;
 ;;; Copyright © 2019, 2020, 2021, 2022 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
 ;;; Copyright © 2019, 2020 Rob Wieringa <rma.wieringa@gmail.com>
-;;; Copyright © 2019, 2020, 2021 Rutger van Beusekom <rutger@dezyne.org>
+;;; Copyright © 2019, 2020, 2021, 2022 Rutger van Beusekom <rutger@dezyne.org>
 ;;; Copyright © 2021 Paul Hoogendijk <paul@dezyne.org>
 ;;; Copyright © 2019 Johri van Eerd <vaneerd.johri@gmail.com>
 ;;;
@@ -42,6 +42,7 @@
   #:use-module (dzn command-line)
   #:use-module (dzn misc)
   #:use-module (dzn parse)
+  #:use-module (dzn parse silence)
   #:export (parse-tree->ast
             annotate-ast))
 
@@ -74,54 +75,119 @@
     (1+ (- pos (car (vector-ref newlines (1- line))))))
 
   (define (file-helper o file-name start-pos)
-    (define (get-location o)
-      (match o
-        (((and (? symbol?) type) body ... ('location pos end))
-         (let* ((ast (helper (cons type body)))
-                (location (helper (last o)))
-                (location (if (and (is-a? ast <root-node>)
-                                   (.location ast)) (clone location #:file-name (.file-name (.location ast)))
-                                   location)))
-           location))
-        (_ #f)))
+
+    (define (make-interface name types-and-events behavior comment)
+      (let* ((types-and-events (helper types-and-events))
+             (comments-t/e types-and-events
+                           (partition (is? <comment-node>) types-and-events))
+             (comment-t/e (and=> (as comments-t/e <pair>) car))
+             (types events (partition (is? <type-node>) types-and-events))
+             (types? (pair? types))
+             (types (match types
+                      ((type tail ...)
+                       (cons (clone type #:comment comment-t/e) tail))
+                      (_
+                       types)))
+             (types (make <types-node> #:elements types))
+             (events (if types? events
+                         (match events
+                           ((event tail ...)
+                            (cons (clone event #:comment comment-t/e) tail))
+                           (_
+                            events))))
+             (events (make <events-node> #:elements events))
+             (behavior (and behavior (annotate-functions (helper behavior)))))
+        (make <interface-node>
+          #:name (helper name)
+          #:comment comment
+          #:types types
+          #:events events
+          #:behavior behavior)))
+
+    (define (comment-ast-list ast comment)
+      (match (.elements ast)
+        ((h t ...)
+         (let ((elements (cons (clone h #:comment comment) t)))
+           (clone ast #:elements elements)))
+        (_
+         ast)))
+
+    (define (comment-statement ast comment)
+      (match ast
+        (($ <compound>)
+         (comment-ast-list ast comment))
+        (_
+         (clone ast #:comment comment))))
+
     (define (helper o)
+      (let* ((ast location comment (strip o))
+             (ast (if (and comment (is-a? ast <root-node>))
+                      (clone ast #:comment comment)
+                      ast)))
+        (values ast location comment)))
+
+    (define (strip o)
+      "Strip comments and locations from the parse tree and add them as fields
+to the AST element."
       (match o
-        ("bool" "bool")
-        ("void" "void")
+        (((and (? symbol?) type) body ... (and ('location pos end) location))
+         (let* ((ast x comment (strip (cons type body)))
+                (location (and location (tree->ast location)))
+                (ast (if (and location (is-a? ast <locationed-node>))
+                         (clone ast #:location location)
+                         ast)))
+           (values ast location comment)))
 
-        ("in" 'in)
-        ("out" 'out)
-        ("inout" 'inout)
+        (((and (? symbol?) type) body ... (and ('comment c ...) comment))
+         (let* ((comment (strip comment))
+                (ast (tree->ast (cons type body) comment)))
+           (cond ((is-a? ast <locationed-node>)
+                  (clone ast #:comment comment))
+                 ((is-a? ast <pair>)
+                  (cons comment ast))
+                 ((is-a? ast <ast-list-node>)
+                  (comment-ast-list ast comment))
+                 (else
+                  ast))
+           (values ast #f comment)))
 
-        ((and (? string?) (? string->number)) (string->number o))
+        (((and (or 'root 'interface 'behavior 'component 'types 'events 'event)
+               type) body ... (and (? string?) string))
+         ;; FIXME: junking non-parsed STRING
+         (values (tree->ast (cons type body)) #f #f))
 
-        ((? string?) o)
+        (((? symbol?) body ...)
+         (values (tree->ast o) #f #f))
 
-        (((and (? symbol?) type) body ... ('comment comment ...))
-         (let ((ast (helper (cons type body))))
-           ast)) ;; TODO ensure (is-a? ast <ast>) is invariant to prevent comment loss
+        ((or "in" "out" "inout")
+         (values (string->symbol o) #f #f))
 
-        (((and (? symbol?) type) body ... ('location pos end))
-         (let ((ast (helper (cons type body))))
-           (if (not (is-a? ast <locationed-node>)) ast
-               (let* ((location (helper (last o)))
-                      (location (if (and (is-a? ast <root-node>) (.location ast))
-                                    (clone location #:file-name (.file-name (.location ast)))
-                                    location)))
-                 (if location (clone ast #:location location)
-                     ast)))))
+        ((and (? string?) (? string->number))
+         (values (string->number o) #f #f))
 
-        (((and (or 'root 'interface 'behavior 'component 'types 'events 'event) type) body ... (? string?))
-         ;; FIXME: junking non-comment-parsed string
-         (helper (cons type body)))
+        ((left (or "||" "&&" "+" "-" "-" "<" "<=" "==" "!=" ">" ">=") right)
+         (values (tree->expression o) #f #f))
+
+        ((or "bool" "void")
+         (values o #f #f))
+
+        ((h ...)
+         (values (filter-map helper o) #f #f))
+
+        (_
+         (values o #f #f))))
+
+    (define* (tree->ast o #:optional comment)
+      (match o
 
         (('comment comment)
-         (make <comment-node> #:comment comment))
+         (make <comment-node> #:string comment))
 
-        (('elements elements ...) (helper elements))
+        (('elements elements ...)
+         (helper elements))
 
         (('import name)
-         (make <import-node> #:name (helper name) #:root (make <root-node>)))
+         (make <import-node> #:name (helper name)))
 
         (('namespace name)
          (make <namespace-node>
@@ -157,40 +223,31 @@
 
         (('interface name)
          (make <interface-node>
-           #:name (helper name)))
+           #:name (helper name)
+           #:comment comment))
 
         (('interface name (and ('types-and-events x ...) types-and-events))
-         (let* ((types-and-events (helper types-and-events))
-                (types events (partition (is? <type-node>) types-and-events)))
-           (make <interface-node>
-             #:name (helper name)
-             #:types (make <types-node> #:elements types)
-             #:events (make <events-node> #:elements events))))
+         (make-interface name types-and-events #f comment))
 
-        (('interface name (and ('behavior x ...) behavior))
-         (let ((behavior (set-recursive (helper behavior))))
-           (make <interface-node>
-             #:name (helper name)
-             #:behavior behavior)))
+        (('interface name (and ('behavior x ...) behavior comment))
+         (make-interface name '() behavior comment))
 
         (('interface name types-and-events behavior)
-         (let* ((types-and-events (helper types-and-events))
-                (types events (partition (is? <type-node>) types-and-events))
-                (behavior (set-recursive (helper behavior))))
-           (make <interface-node>
-             #:name (helper name)
-             #:types (make <types-node> #:elements types)
-             #:events (make <events-node> #:elements events)
-             #:behavior behavior)))
+         (make-interface name types-and-events behavior comment))
 
         (('types-and-events types-and-events ...)
          (helper types-and-events))
 
         (('event direction type name formals)
-         (make <event-node>
-           #:name (helper name)
-           #:signature (make <signature-node> #:type.name (helper type) #:formals (helper formals) #:location (get-location type))
-           #:direction (helper direction)))
+         (let* ((type location comment (helper type))
+                (signature (make <signature-node>
+                             #:type.name type
+                             #:formals (helper formals)
+                             #:location location)))
+           (make <event-node>
+             #:name (helper name)
+             #:signature signature
+             #:direction (helper direction))))
 
         (('formals formal) (make <formals-node> #:elements (list (helper formal))))
         (('formals formals ...) (make <formals-node> #:elements (helper formals)))
@@ -211,20 +268,23 @@
 
         (('component name ports)
          (make <foreign-node>
+           #:comment comment
            #:name (helper name)
            #:ports (helper ports)))
 
         (('component name ports (and ('behavior elements ...) behavior))
          (make <component-node>
+           #:comment comment
            #:name (helper name)
            #:ports (helper ports)
-           #:behavior (set-recursive (helper behavior))))
+           #:behavior (annotate-functions (helper behavior))))
 
         (('component name ports ('system instances-and-bindings rest ...))
          (let* ((instances-and-bindings (helper instances-and-bindings))
                 (instances
                  bindings (partition (is? <instance-node>) instances-and-bindings)))
            (make <system-node>
+             #:comment comment
              #:name (helper name)
              #:ports (helper ports)
              #:instances (make <instances-node> #:elements instances)
@@ -316,10 +376,13 @@
              #:elements body)))
 
         (('on triggers)
-         (make <on-node> #:triggers (helper triggers)))
+         (make <on-node> #:comment comment #:triggers (helper triggers)))
 
         (('on triggers statement)
-         (make <on-node> #:triggers (helper triggers) #:statement (helper statement)))
+         (make <on-node>
+           #:comment comment
+           #:triggers (helper triggers)
+           #:statement (helper statement)))
 
         (('action-or-call action-or-call)
          (helper action-or-call))
@@ -365,24 +428,36 @@
         (('type-name name) (helper name))
         (('event-name name) (helper name))
 
-        (('interface-action event) (make <action-node> #:event.name (helper event)))
+        (('interface-action event)
+         (make <action-node>
+           #:event.name (helper event)))
 
-        (('illegal) (make <illegal-node>))
+        (('illegal)
+         (make <illegal-node>))
 
-        (('action (port event)) (make <action-node> #:port.name (helper port) #:event.name (helper event)))
+        (('action port event)
+         (make <action-node>
+           #:port.name (helper port)
+           #:event.name (helper event)))
 
-        (('action (port event) arguments) (make <action-node> #:port.name (helper port) #:event.name (helper event) #:arguments (helper arguments)))
+        (('action port event arguments)
+         (make <action-node>
+           #:port.name (helper port)
+           #:event.name (helper event)
+           #:arguments (helper arguments)))
 
-        (('action port event) (make <action-node> #:port.name (helper port) #:event.name (helper event)))
+        (('assign var expression)
+         (make <assign-node>
+           #:variable.name (.name (helper var))
+           #:expression (helper expression)))
 
-        (('action port event arguments) (make <action-node> #:port.name (helper port) #:event.name (helper event) #:arguments (helper arguments)))
-
-        (('assign var expression) (make <assign-node>
-                                    #:variable.name (.name (helper var))
-                                    #:expression (helper expression)))
-
-        (('behavior-statement statement ...)
-         (map helper statement))
+        (('behavior-statements statement ...)
+         (let ((statements (map helper statement)))
+           (match statements
+             ((h t ...)
+              (cons (comment-statement h comment) t))
+             (_
+              statements))))
 
         (('behavior-compound statement)
          (make <compound-node> #:elements (helper statement)))
@@ -415,6 +490,14 @@
            #:function.name (helper function)
            #:arguments (helper arguments)))
 
+        (('defer statement)
+         (make <defer-node>
+           #:statement (helper statement)))
+
+        (('defer arguments statement)
+         (make <defer-node>
+           #:arguments (helper arguments)
+           #:statement (helper statement)))
 
         (('arguments argument) (make <arguments-node> #:elements (list(helper argument))))
         (('arguments argument ...) (make <arguments-node> #:elements (helper argument)))
@@ -436,10 +519,11 @@
         (('field-test ('unknown-identifier identifier _ ...) field) (make <field-test-node> #:variable.name identifier #:field (helper field)))
 
         (('function type name formals statement)
-         (let ((signature (make <signature-node>
-                            #:type.name (helper type)
-                            #:formals (helper formals)
-                            #:location (get-location type))))
+         (let* ((type location comment (helper type))
+                (signature (make <signature-node>
+                             #:type.name type
+                             #:formals (helper formals)
+                             #:location location)))
            (make <function-node>
              #:name (helper name)
              #:signature signature
@@ -450,10 +534,12 @@
 
         (('guard expression)
          (make <guard-node>
+           #:comment comment
            #:expression (helper expression)))
 
         (('guard expression statement)
          (make <guard-node>
+           #:comment comment
            #:expression (helper expression)
            #:statement (helper statement)))
 
@@ -538,18 +624,6 @@
         (('expression expression) (helper expression))
         (('expression expression ...) (helper expression))
         (('group expression) (make <group-node> #:expression (helper expression)))
-        ((left "||" right) (make <or-node> #:left (helper left) #:right (helper right)))
-        ((left "&&" right) (make <and-node> #:left (helper left) #:right (helper right)))
-
-        ((left "+" right) (make <plus-node> #:left (helper left) #:right (helper right)))
-        ((left "-" right) (make <minus-node> #:left (helper left) #:right (helper right)))
-        ((left "-" right) (make <minus-node> #:left (helper left) #:right (helper right)))
-        ((left "<" right) (make <less-node> #:left (helper left) #:right (helper right)))
-        ((left "<=" right) (make <less-equal-node> #:left (helper left) #:right (helper right)))
-        ((left "==" right) (make <equal-node> #:left (helper left) #:right (helper right)))
-        ((left "!=" right) (make <not-equal-node> #:left (helper left) #:right (helper right)))
-        ((left ">"  right) (make <greater-node> #:left (helper left) #:right (helper right)))
-        ((left ">=" right) (make <greater-equal-node> #:left (helper left) #:right (helper right)))
         (('not expression) (make <not-node> #:expression (helper expression)))
 
         (('literal "true") (make <literal-node> #:value "true"))
@@ -562,12 +636,22 @@
                 (column (column-number line pos))
                 (end-line (line-number end))
                 (end-column (column-number end-line end)))
-           (make <location-node> #:file-name file-name #:line (1+ (- line start-line)) #:column column #:end-line (1+ (- end-line start-line)) #:end-column end-column #:offset (- pos start-pos) #:length (- end pos))))
+           (make <location-node> #:file-name file-name #:line (1+ (- line start-line)) #:column column #:end-line (1+ (- end-line start-line)) #:end-column end-column #:offset (- pos start-pos) #:length (- end pos))))))
 
-        ((? (is? <ast>)) o)
+    (define (tree->expression o)
+      (match o
+        ((left "||" right) (make <or-node> #:left (helper left) #:right (helper right)))
+        ((left "&&" right) (make <and-node> #:left (helper left) #:right (helper right)))
+        ((left "+" right) (make <plus-node> #:left (helper left) #:right (helper right)))
+        ((left "-" right) (make <minus-node> #:left (helper left) #:right (helper right)))
+        ((left "-" right) (make <minus-node> #:left (helper left) #:right (helper right)))
+        ((left "<" right) (make <less-node> #:left (helper left) #:right (helper right)))
+        ((left "<=" right) (make <less-equal-node> #:left (helper left) #:right (helper right)))
+        ((left "==" right) (make <equal-node> #:left (helper left) #:right (helper right)))
+        ((left "!=" right) (make <not-equal-node> #:left (helper left) #:right (helper right)))
+        ((left ">"  right) (make <greater-node> #:left (helper left) #:right (helper right)))
+        ((left ">=" right) (make <greater-equal-node> #:left (helper left) #:right (helper right)))))
 
-        ((h ...) (filter-map helper o))
-        (_ (format #f "LITERAL: \"~s\"" o))))
     (helper o))
 
   (let* ((root-node (file-helper o file-name 0))
@@ -575,16 +659,22 @@
     (tree-map make-namespaces root)))
 
 (define-method (set-recursive (o <behavior>))
+  (define (mark-recursive f)
+    (if (ast:recursive? f) (clone f #:recursive? #t)
+        f))
   (let* ((functions (.functions o))
          (function-list (.elements functions))
-         (function-list (map (lambda (f) (if (ast:recursive? f) (clone f #:recursive #t) f))
-                             function-list))
+         (function-list (map mark-recursive function-list))
          (functions (clone functions #:elements function-list)))
     (clone o #:functions functions)))
 
-(define-method (set-recursive (o <behavior-node>))
-  (.node (set-recursive (make <behavior> #:node o))))
+(define-method (set-noisy (o <behavior>))
+  (mark-noisy o))
 
+(define-method (annotate-functions (o <behavior-node>))
+  (let* ((o (set-recursive (make <behavior> #:node o)))
+         (o (set-noisy o)))
+   (.node o)))
 
 (define-method (make-namespaces (o <ast>))
   o)
