@@ -32,7 +32,6 @@
   #:use-module (ice-9 rdelim)
 
   #:use-module (dzn peg)
-  #:use-module (dzn parse ast)
 
   #:re-export (%peg:locations?
                %peg:skip?
@@ -48,7 +47,9 @@
 (define-skip-parser peg-eof none (not-followed-by peg-any))
 (define-skip-parser peg-eol none (or "\f" "\n" "\r" "\v"))
 (define-skip-parser peg-ws none (or " " "\t"))
-(define-skip-parser peg-line all (and "//" (* (and (not-followed-by peg-eol) peg-any))))
+(define-skip-parser peg-line all
+  (and "//" (* (and (not-followed-by peg-eol) peg-any))
+       (expect (or "\n" "\r\n" peg-eof))))
 (define-skip-parser peg-block-strict all (and "/*" (* (or peg-block (and (not-followed-by "*/") peg-any))) (expect "*/")))
 (define-skip-parser peg-skip all (* (or peg-ws peg-eol peg-line peg-block-strict)))
 
@@ -111,6 +112,8 @@ SKIP < !IMPORT . 'import'*")
   (define variable-stack '())
 
   (define (-enter-frame- str len pos)
+    (when (%peg:debug?)
+      (format (current-error-port) "push: ~a\n" variable-stack))
     (set! variable-stack (cons (if (null? variable-stack) '() (car variable-stack))
                                variable-stack))
     (list pos '()))
@@ -118,6 +121,8 @@ SKIP < !IMPORT . 'import'*")
 
   (define (-exit-frame- str len pos)
     (set! variable-stack (if (null? variable-stack) '()  (cdr variable-stack)))
+    (when (%peg:debug?)
+      (format (current-error-port) "pop: ~a\n" variable-stack))
     (list pos '()))
   (define-skip-parser exit-frame none -exit-frame-)
 
@@ -215,7 +220,10 @@ compound <-- BRACE-OPEN enter-frame (statement / !unknown-identifier)#* BRACE-CL
 
 imperative-statement <- variable / assign / if-statement / illegal /
                         return / skip-statement / compound /
-                        reply / action-or-call / interface-action SEMICOLON#
+                        reply / defer / action-or-call /
+                        interface-action SEMICOLON#
+
+  defer <-- DEFER arguments? imperative-statement
 
   interface-action <-- is-event
 
@@ -246,7 +254,7 @@ or-expression <- and-expression OR or-expression# / and-expression
 and-expression <- compare-expression AND and-expression# / compare-expression
 compare-expression <- plus-min-expression !LEFT-ARROW COMPARE plus-min-expression# / plus-min-expression
 plus-min-expression <- not-expression (PLUS / MINUS) not-expression# / not-expression
-not-expression <- not / group / dollars / (!var !is-port enum-literal / literal / var !DOT / action / call / interface-action / field-test / unknown-identifier)
+not-expression <- not / group / dollars / (!var !is-port enum-literal / literal / call / var !DOT / action / interface-action / field-test / unknown-identifier)
 not <-- NOT not-expression#
 enum-literal <-- global? scope name
 field-test <-- !is-port var DOT# name#
@@ -307,6 +315,7 @@ BEHAVIOR            <  'behavior' ![a-zA-Z_0-9] / 'behaviour' ![a-zA-Z_0-9]
 BLOCKING            <  'blocking' ![a-zA-Z_0-9]
 BOOL                <- 'bool' ![a-zA-Z_0-9]
 COMPONENT           <  'component' ![a-zA-Z_0-9]
+DEFER               <  'defer' ![a-zA-Z_0-9]
 ELSE                <  'else' ![a-zA-Z_0-9]
 ENUM                <  'enum' ![a-zA-Z_0-9]
 EXTERN              <  'extern' ![a-zA-Z_0-9]
@@ -340,6 +349,7 @@ KEYWORD <
   / 'behaviour' ![a-zA-Z_0-9]
   / 'blocking' ![a-zA-Z_0-9]
   / 'component' ![a-zA-Z_0-9]
+  / 'defer' ![a-zA-Z_0-9]
   / 'else' ![a-zA-Z_0-9]
   / 'enum' ![a-zA-Z_0-9]
   / 'extern' ![a-zA-Z_0-9]
