@@ -1,6 +1,6 @@
 ;;; Dezyne --- Dezyne command line tools
 ;;;
-;;; Copyright © 2017, 2018, 2019, 2020, 2021, 2023 Janneke Nieuwenhuizen <janneke@gnu.org>
+;;; Copyright © 2017, 2018, 2019, 2020, 2021, 2022 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
 ;;; Copyright © 2018 Rob Wieringa <rma.wieringa@gmail.com>
 ;;;
 ;;; This file is part of Dezyne.
@@ -23,12 +23,13 @@
 ;;; Code:
 
 (define-module (dzn script)
+  #:use-module (ice-9 documentation)
   #:use-module (ice-9 getopt-long)
+  #:use-module (ice-9 match)
   #:use-module (ice-9 poe)
   #:use-module (system repl error-handling)
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-26)
-  #:use-module (srfi srfi-71)
   #:use-module (dzn config)
   #:use-module (dzn command-line)
   #:use-module (dzn misc)
@@ -41,35 +42,54 @@
 	  '((debug (single-char #\d))
             (help (single-char #\h))
 	    (skip-wfc (single-char #\p))
+	    (transform (single-char #\t) (value #t))
 	    (verbose (single-char #\v))
 	    (version (single-char #\V))))
 	 (options (getopt-long args option-spec
 		               #:stop-at-first-non-option #t))
+	 (verbose? (option-ref options 'verbose #f))
 	 (help? (option-ref options 'help #f))
 	 (files (option-ref options '() '()))
 	 (usage? (and (not help?) (null? files)))
 	 (version? (option-ref options 'version #f)))
 
     (define (list-commands dir)
-      (let* ((uninstalled? (getenv "DZN_UNINSTALLED"))
-             (ext path (if uninstalled? (values ".scm" %load-path)
-                           (values ".go" %load-compiled-path))))
-       (map (cut basename <> ext)
-            (filter
-             (cute string-contains <> dir)
-             (append-map
-              (cute list-directory <>
-                    (cute string-suffix? ext <>))
-              (filter directory-exists?
-                      (map (cute string-append <> "/dzn/commands/")
-                           path)))))))
+      (map (cut basename <> ".go")
+           (filter
+            (cute string-contains <> dir)
+            (append-map
+             (cute list-directory <>
+                   (cute string-suffix? ".go" <>))
+             (filter directory-exists?
+                     (map (cute string-append <> "/dzn/commands/")
+                          %load-compiled-path))))))
 
     (when version?
       (show-version-and-exit))
     (when (or help? usage?)
       (let* ((port (if usage? (current-error-port) (current-output-port)))
              (commands (list-commands "/dzn/commands/"))
-             (commands (sort (delete-duplicates commands string=?) string<)))
+             (commands (sort (delete-duplicates commands string=?) string<))
+             (transform (resolve-interface `(dzn transform)))
+             (symbol.var->string.doc
+              (match-lambda
+                ((symbol . var)
+                 (cons (string-append "  " (symbol->string symbol))
+                       (object-documentation (variable-ref var))))))
+             (transformations (map symbol.var->string.doc
+                                   (module-map cons transform)))
+             (string< (lambda (a b) (string< (car a) (car b))) )
+             (transformations (sort transformations string<))
+             (prologue-length (map (compose string-length car) transformations))
+             (tabs (ceiling (/ (apply max prologue-length) 8)))
+             (string.doc->string
+              (match-lambda
+                ((string . doc)
+                 (let* ((prologue (string-append string ":"))
+                        (block-format (cute block-format <> prologue tabs 80))
+                        (doc (and=> doc block-format)))
+                   (if (not doc) string
+                       (string-append doc "\n")))))))
         (format port "\
 Usage: dzn [OPTION]... COMMAND [COMMAND-ARGUMENT...]
   -d, --debug            enable debug ouput
@@ -77,12 +97,19 @@ Usage: dzn [OPTION]... COMMAND [COMMAND-ARGUMENT...]
   -p, --skip-wfc         use plain PEG, skip well-formedness checking
   -v, --verbose          be more verbose, show progress
   -V, --version          display version
+  -t, --transform=TRANS  use transformation TRANS
 
 Commands:~a
 
+Transformations: ~a
+
 Use \"dzn COMMAND --help\" for command-specific information.
 "
-                (string-join commands "\n  " 'prefix)))
+                (string-join commands "\n  " 'prefix)
+                (if verbose? (string-join
+                              (map string.doc->string transformations)
+                              "\n" 'prefix)
+                  "Use dzn --help --verbose to list transformations.")))
       (exit (or (and usage? EXIT_OTHER_FAILURE) EXIT_SUCCESS)))
     options))
 
