@@ -37,7 +37,6 @@
   #:use-module (dzn misc)
   #:use-module (dzn parse)
   #:use-module (dzn shell-util)
-  #:use-module (dzn ast)
   #:use-module (dzn vm ast)
   #:use-module (dzn vm evaluate)
   #:use-module (dzn vm goops)
@@ -185,18 +184,26 @@ format."
   "Check TRACE for a V-fork with respect to PORT.  If TRACE contains
 actions to a provides port other than PORT, mark the trace as
 <fork-error>, otherwise return false."
-  (let ((action-step (list-index (cute action-other-provides-port? port <>)
-                                 trace)))
-    (and action-step
+  (let* ((pc (find (conjoin
+                    .trigger
+                    (compose .port .trigger)
+                    (compose (is? <on>) .statement)
+                    (compose ast:provides? .port .trigger))
+                   trace))
+         (port (or (and pc (.port (.trigger pc))) port))
+         (action-step (list-index (cute action-other-provides-port? port <>)
+                                  trace)))
+    (and pc
+         action-step
          (let* ((trace (drop trace action-step))
-                (pc (car trace))
+                (action-pc (car trace))
+                (action (.statement action-pc))
                 (pc (clone pc
                            #:previous #f
                            #:status (make <fork-error>
-                                      #:ast (.statement pc)
+                                      #:ast action
                                       #:message "non-compliance"))))
-           (and (ast:provides? (.trigger pc))
-                (list (cons pc trace)))))))
+           (list (cons pc trace))))))
 
 (define (check-requires-provides-fork trace)
   "Check TRACE for a Y-fork.  If TRACE contains actions to more than one
@@ -233,7 +240,7 @@ Return a list of traces, possibly marked with <compliance-error>."
            (instance (.container r:component-port))
            (component-trigger (trigger->component-trigger trigger))
            (at (list-index
-                (if (ast:provides? trigger)
+                (if (ast:provides? component-trigger)
                     (conjoin
                      (compose (cute eq? <> instance) .instance)
                      (compose (is? <initial-compound>) .statement)
@@ -246,12 +253,13 @@ Return a list of traces, possibly marked with <compliance-error>."
                 trace))
            (trace (if (not at) trace
                       (list-head trace (1+ at))))
-           (trace (if (ast:provides? trigger) trace
+           (trace (if (ast:provides? component-trigger) trace
                       (filter (compose (negate (is? <trigger-return>)) .statement) trace))))
       trace))
 
   (let* ((component ((compose .type .ast) (%sut)))
          (trigger (and (string? event)
+                       (not (equal? event "<defer>"))
                        (clone (string->trigger event) #:parent component)))
          (blocking? (find (compose pair? .blocked) trace))
          (sut-trace (if (or (not trigger) (not blocking?)) trace
@@ -284,155 +292,163 @@ Return a list of traces, possibly marked with <compliance-error>."
               (append-map (cute run-to-completion ipc <>) modeling-names))))
 
         (define (run-provides-port pc event)
-          (%debug "run-provides-port... ~a\n" event)
+          (%debug "run-provides-port... ~a ~a\n" port-instance event)
+          (%debug "  pc: ~a\n" pc)
           (parameterize ((%sut port-instance)
                          (%exploring? #t))
             (run-to-completion pc event)))
 
         (%debug "check-provides-compliance... ~s: ~a [~a]\n" port-name event port-event)
-        (let* ((interface ((compose .type .ast) port-instance))
-               (ipc (clone pc #:previous #f #:trail '() #:status #f #:statement #f))
-               (ipc (reset-reply pc port-instance))
+        (let* ((ipc (clone pc #:previous #f #:trail '() #:status #f #:statement #f))
+               (ipc (reset-reply ipc port-instance))
                (port-traces (if port-event (run-provides-port ipc port-event)
-                                (run-provides-modeling ipc port-instance))))
+                                (run-provides-modeling ipc port-instance)))
+               (port-prefix (format #f "~a." port-name))
+               (sut-trail (filter (compose (disjoin
+                                            (cute equal? <> "<illegal>")
+                                            (cute string-prefix? port-prefix <>))
+                                           cdr)
+                                  sut-trail))
+               (blocked? (and (pair? trace)
+                              (or (pair? (.blocked (car trace)))
+                                  (find blocked-on-boundary? trace)))))
 
-          (let* ((port-prefix (format #f "~a." port-name))
-                 (sut-trail (filter (compose (disjoin (cut equal? <> "illegal")
-                                                      (cut string-prefix? port-prefix <>))
-                                             cdr)
-                                    sut-trail))
-                 (blocked? (and (pair? trace)
-                                (or (pair? (.blocked (car trace)))
-                                    (find blocked-on-boundary? trace)))))
+          (define (port-trace->trail trace)
+            (parameterize ((%sut port-instance)) (trace->trail trace)))
 
-            (define (port-trace->trail trace)
-              (parameterize ((%sut port-instance)) (trace->trail trace)))
+          (define (first-non-match port-trace)
+            (define (non-matching-pair? a b)
+              (and (not (equal? (cdr a) ((compose last (cut string-split <> #\.) cdr) b))) (cons a b)))
 
-            (define (first-non-match port-trace)
-              (define (non-matching-pair? a b)
-                (and (not (equal? (cdr a) ((compose last (cut string-split <> #\.) cdr) b))) (cons a b)))
+            (let* ((port-trail (port-trace->trail port-trace))
+                   (truncate? (and provides-trigger? blocked?))
+                   (port-trail (if (not truncate?) port-trail
+                                   ;; Check prefix only as long as trace is blocked
+                                   (list-head port-trail (min (length port-trail)
+                                                              (length sut-trail)))))
+                   (foo (%debug "     port trail : ~s\n" (map cdr port-trail)))
+                   (port-name ((compose .name .ast) port-instance))
+                   (events (map (compose last (cut string-split <> #\.) cdr) sut-trail))
+                   (foo (%debug "      sut trail : ~s\n\n" events)))
+              (or (any non-matching-pair? port-trail sut-trail)
+                  (let ((port-length (length port-trail))
+                        (sut-length (length sut-trail)))
+                    (cond ((< port-length sut-length) (cons '(#f) (list-ref sut-trail port-length)))
+                          ((> port-length sut-length) (list (list-ref port-trail sut-length) #f))
+                          (else #f))))))
 
-              (let* ((port-trail (port-trace->trail port-trace))
-                     (truncate? (and provides-trigger? blocked?))
-                     (port-trail (if (not truncate?) port-trail
-                                     ;; Check prefix only as long as trace is blocked
-                                     (list-head port-trail (min (length port-trail)
-                                                                (length sut-trail)))))
-                     (foo (%debug "     port trail : ~s\n" (map cdr port-trail)))
-                     (port-name ((compose .name .ast) port-instance))
-                     (events (map (compose last (cut string-split <> #\.) cdr) sut-trail))
-                     (foo (%debug "      sut trail : ~s\n\n" events)))
-                (or (any non-matching-pair? port-trail sut-trail)
-                    (let ((port-length (length port-trail))
-                          (sut-length (length sut-trail)))
-                      (cond ((< port-length sut-length) (cons '(#f) (list-ref sut-trail port-length)))
-                            ((> port-length sut-length) (list (list-ref port-trail sut-length) #f))
-                            (else #f))))))
+          (define (port-acceptance-equal? a b)
+            (and (equal? (and=> (caar a) trigger->string)
+                         (and=> (caar b) trigger->string))
+                 (equal? (cadr a) (cadr b))))
 
-            (define (port-acceptance-equal? a b)
-              (and (equal? (and=> (caar a) trigger->string)
-                           (and=> (caar b) trigger->string))
-                   (equal? (cadr a) (cadr b))))
+          (define (event-on-trail? event-name trace)
+            (let* ((trail (port-trace->trail trace))
+                   (trail (map cdr trail)))
+              (member event-name trail)))
 
-            (define (event-on-trail? event-name trace)
-              (let* ((trail (port-trace->trail trace))
-                     (trail (map cdr trail)))
-                (member event-name trail)))
+          (when (> (dzn:debugity) 0)
+            (%debug "sut-trail:~s\n" (map cdr sut-trail))
+            (%debug "port-traces[~s]:\n" (length port-traces))
+            (parameterize ((%sut port-instance))
+              (display-trails port-traces)))
 
-            (when (> (dzn:debugity) 0)
-              (%debug "sut-trail:~s\n" (map cdr sut-trail))
-              (%debug "port-traces[~s]:\n" (length port-traces))
-              (parameterize ((%sut port-instance))
-                (display-trails port-traces)))
-
-            (let ((port-traces
-                   non-compliances
-                   (partition (if (not (%compliance-check?)) identity
-                                  (negate first-non-match)) port-traces)))
-              (cond
-               ((and (pair? trace)
-                     (.status (car trace))
-                     (not (is-a? (.status (car trace)) <match-error>))
-                     (pair? (append port-traces non-compliances)))
-                (%debug "  exit 0\n")
-                (let* ((pc (car trace))
-                       (status (.status pc))
-                       (trace (rewrite-trace-head (cut clone <> #:status #f #:statement #f) trace))
-                       (trace (zip trigger trace (car (append port-traces non-compliances))))
-                       (trace (cons (car trace) trace))
-                       (trace (rewrite-trace-head (cut clone <> #:status status) trace)))
+          (let ((port-traces
+                 non-compliances
+                 (partition (negate first-non-match) port-traces)))
+            (cond
+             ((and (pair? trace)
+                   (.status (car trace))
+                   (not (is-a? (.status (car trace)) <match-error>))
+                   (pair? (append port-traces non-compliances)))
+              (%debug "  exit 0\n")
+              (let* ((trace (rewrite-trace-head (cut clone <> #:statement #f) trace))
+                     (trace (zip trigger trace (car (append port-traces non-compliances)))))
+                (list trace)))
+             ((and (pair? port-traces)
+                   (pair? trace))
+              (%debug "  exit 1\n")
+              (let* ((port-pcs (map (compose (cut clone pc #:state <>) .state car) port-traces))
+                     (traces (map (lambda (port-pc)
+                                    (cons (set-state (car trace) (get-state port-pc port-instance))
+                                          (cdr trace)))
+                                  port-pcs)))
+                (map (cute zip trigger <> <>) traces port-traces)))
+             ((and (%compliance-check?)
+                   (null? non-compliances)
+                   (not blocking?)
+                   (null? port-traces)
+                   (pair? sut-trail))
+              (%debug "  exit 2\n")
+              (let ((status (make <compliance-error>
+                              #:message "non-compliance"
+                              #:component-acceptance (caar sut-trail)
+                              #:port port-instance)))
+                (list (rewrite-trace-head (cut clone <> #:status status) trace))))
+             ((null? non-compliances)
+              (%debug "  exit 3\n")
+              (if (null? trace) '()
                   (list trace)))
-               ((and (pair? port-traces)
-                     (pair? trace))
-                (%debug "  exit 1\n")
-                (let* ((port-pcs (map (compose (cut clone pc #:state <>) .state car) port-traces))
-                       (traces (map (lambda (port-pc)
-                                      (cons (set-state (car trace) (get-state port-pc port-instance))
-                                            (cdr trace)))
-                                    port-pcs)))
-                  (map (cute zip trigger <> <>) traces port-traces)))
-               ((and (%compliance-check?)
-                     (null? non-compliances)
-                     (not blocking?)
-                     (null? port-traces)
-                     (pair? sut-trail))
-                (%debug "  exit 2\n")
-                (let ((status (make <compliance-error>
-                                #:message "non-compliance"
-                                #:component-acceptance (caar sut-trail)
-                                #:port port-instance)))
-                  (list (rewrite-trace-head (cut clone <> #:status status) trace))))
-               ((null? non-compliances)
-                (%debug "  exit 3\n")
-                (if (null? trace) '()
-                    (list trace)))
-               ((and (not port-event)
-                     (null? sut-trail)
-                     (pair? trace))
-                (%debug "  exit 4\n")
-                (list trace))
-               (else
-                (%debug "  exit 5\n")
-                (let* ((port-acceptances (map first-non-match non-compliances))
-                       (port-acceptances (delete-duplicates port-acceptances port-acceptance-equal?))
-                       (component-acceptance (and (pair? trace)
-                                                  (or (and (pair? port-acceptances)
-                                                           (cadar port-acceptances))
-                                                      (and=> (.status pc) .ast)
-                                                      (trigger->component-trigger trigger))))
-                       (other-port-instances (runtime:other-port port-instance))
-                       (instance (.container other-port-instances))
-                       (component-acceptance
-                        (if (and component-acceptance (is-a? (%sut) <runtime:system>))
-                            (trigger->system-trigger instance component-acceptance)
-                            component-acceptance))
-                       (port-acceptances (make <acceptances> #:elements (map caar port-acceptances)))
-                       (compliance-trigger
-                        (and (null? sut-trail)
-                             (not (any (cute event-on-trail? (.event.name trigger) <>)
-                                       non-compliances))
-                             trigger))
-                       (pc (clone pc
-                                  #:previous #f
-                                  #:status (make <compliance-error>
-                                             #:message "non-compliance"
-                                             #:component-acceptance component-acceptance
-                                             #:port port-instance
-                                             #:port-acceptance port-acceptances
-                                             #:trigger compliance-trigger))))
-                  (if (null? trace) (list (cons pc (car non-compliances)))
-                      (let* ((tail (cdr trace))
-                             (trace (cons pc tail)))
-                        (list (zip trigger trace (car non-compliances)))))))))))))
+             ((and (not port-event)
+                   (null? sut-trail)
+                   (pair? trace))
+              (%debug "  exit 4\n")
+              (list trace))
+             ((%compliance-check?)
+              (%debug "  exit 5\n")
+              (let* ((port-acceptances (map first-non-match non-compliances))
+                     (port-acceptances (delete-duplicates port-acceptances port-acceptance-equal?))
+                     (component-acceptance (and (pair? trace)
+                                                (or (and (pair? port-acceptances)
+                                                         (cadar port-acceptances))
+                                                    (and=> (.status pc) .ast)
+                                                    (trigger->component-trigger trigger))))
+                     (other-port-instances (runtime:other-port port-instance))
+                     (instance (.container other-port-instances))
+                     (component-acceptance
+                      (if (and component-acceptance (is-a? (%sut) <runtime:system>))
+                          (trigger->system-trigger instance component-acceptance)
+                          component-acceptance))
+                     (port-acceptances (make <acceptances> #:elements (map caar port-acceptances)))
+                     (compliance-trigger
+                      (and trigger
+                           (null? sut-trail)
+                           (not (any (cute event-on-trail? (.event.name trigger) <>)
+                                     non-compliances))
+                           trigger))
+                     (pc (clone pc
+                                #:previous #f
+                                #:status (make <compliance-error>
+                                           #:message "non-compliance"
+                                           #:component-acceptance component-acceptance
+                                           #:port port-instance
+                                           #:port-acceptance port-acceptances
+                                           #:trigger compliance-trigger))))
+                (if (null? trace) (list (cons pc (car non-compliances)))
+                    (let* ((tail (cdr trace))
+                           (trace (cons pc tail)))
+                      (list (zip trigger trace (car non-compliances)))))))
+             (else
+              (%debug "  exit 6\n")
+              (let* ((port-trace (car non-compliances))
+                     (port-state (get-state (last port-trace)))
+                     (port-trace
+                      (rewrite-trace-head
+                       (cut set-state <> port-state)
+                       port-trace)))
+                (list (zip trigger trace (car non-compliances))))))))))
 
-    (if port (or (and (> (length (ast:provides-port* component)) 1)
+    (if port (or (and (%compliance-check?)
+                      (> (length (ast:provides-port* component)) 1)
                       (is-a? (%sut) <runtime:component>)
                       (check-provides-fork port trace))
                  (check-compliance port (list trace)))
         (let ((ports (ast:provides-port* component)))
-          (or (and (> (length ports) 1)
+          (or (and (%compliance-check?)
+                   (> (length ports) 1)
                    (is-a? (%sut) <runtime:component>)
-                   (check-requires-provides-fork trace))
+                   (or (check-provides-fork #f trace)
+                       (check-requires-provides-fork trace)))
               (fold check-compliance (list trace) ports))))))
 
 (define* (check-provides-compliance+ pc event trace)
@@ -441,8 +457,7 @@ released by a requires event, also rerun check-provides-compliance for
 the full trace, i.e., starting from the initial blocking provides
 event.  This ensures proper of zipping the port trace, including the
 port return."
-  (let* ((skip? (or (return-trigger? event)
-                    (blocked-on-boundary? (car trace) event)))
+  (let* ((skip? (blocked-on-boundary? (car trace) event))
          (traces (if skip? (list trace)
                      (check-provides-compliance pc event trace)))
          (pc (car trace))
@@ -576,17 +591,19 @@ port return."
     (let* ((boundary (filter runtime:boundary-port? (%instances)))
            (traces (append-map (cute port->label-traces pc <>) boundary))
            (fake-traces (list (list pc)))
+           (defer-traces (map (cute cons <> fake-traces) (defer-labels pc)))
            (return-traces (map (cute cons <> fake-traces) (return-labels pc)))
            (rtc-traces (map (cute cons <> fake-traces) (rtc-labels pc))))
-      (append traces return-traces rtc-traces)))
+      (append traces defer-traces return-traces rtc-traces)))
 
   (if (is-a? (%sut) <runtime:system>) (system-event-traces-alist pc)
       (event-traces-alist pc)))
 
 (define-method (event-traces-alist (pcs <list>))
-  (let ((pcs (map (cute clone <> #:status #f) pcs)))
-    (merge-alist-list
-     (map event-traces-alist pcs))))
+  (let* ((pcs (map (cute clone <> #:status #f) pcs))
+         (alists (map event-traces-alist pcs)))
+    (if (null? alists) '()
+        (merge-alist-list alists))))
 
 (define-method (eligible-labels event-traces-alist)
   (let* ((eligible-traces
@@ -619,7 +636,9 @@ TRACES."
 
   (let* ((pcs (map car traces))
          (blocked-sets (map blocked+collateral-ports pcs))
-         (blocked-port-names (apply lset-intersection equal? blocked-sets)))
+         (blocked-port-names
+          (if (null? blocked-sets) '()
+              (apply lset-intersection equal? blocked-sets))))
     (filter (disjoin
              rtc-event?
              return-trigger?
@@ -772,7 +791,9 @@ or false."
 
 (define-method (run-sut (pc+blocked-trace <list>) event)
   (let* ((pc (car pc+blocked-trace))
-         (pc (reset-replies pc)))
+         (blocked-trace (cdr pc+blocked-trace))
+         (pc (reset-replies pc))
+         (pc (prune-defer pc)))
     (%debug "run-sut pc: ~s\n" pc)
     (%debug "     event: ~s\n" event)
     (match event
@@ -799,7 +820,14 @@ or false."
        (list (cons (clone pc #:async '()) pc+blocked-trace)))
       ((? (const (pair? (.async pc))))
        (let ((traces (flush-async pc)))
-         (check-provides-compliance* pc #f traces)))
+         (check-provides-compliance* pc event traces)))
+      ((? (const (and (pair? (.defer pc)) (blocked-on-boundary? pc))))
+       (list (cons (clone pc #:defer '()) pc+blocked-trace)))
+      ((? (const (and (pair? (.defer pc)) (not (%strict?)))))
+       (let* ((traces (flush-defer pc))
+              (traces (if (null? blocked-trace) traces
+                          (extend-trace blocked-trace (const traces)))))
+         (check-provides-compliance* pc event traces)))
       (_
        (when (not (eq? (switch-context pc) pc))
          (%debug "<eot> with switchable, non-rtc pc\n"))
@@ -822,7 +850,7 @@ optional labels only and stop when observable event seen."
                             #:labels (const '("inevitable" "optional"))))))
     (when (%debug?)
       (parameterize ((%modeling? #t))
-        ((@ (ice-9 pretty-print) pretty-print)
+        (pretty-print
          (debug:lts->alist pc->state-number lts) (current-error-port))))
     lts))
 
@@ -1004,33 +1032,36 @@ status."
        ;; In the failures model, refusals can only occur when the
        ;; component LTS is stable.  When the component LTS is not
        ;; stable, that means outgoing tau events: no refusals.  A
-       ;; component LTS is stable when all requires port LTSs are
-       ;; stable.  When the a provides port's modeling-LTS is not
-       ;; stable, the refusals set consists of all its observable
-       ;; events.
-       (and (= (length from-pcs) 1)
-            (null? (.external-q (car from-pcs)))
-            (requires-ports-stable? component pc)
-            (let* ((ports (ast:provides-port* component))
-                   (instable (find (negate (port-lts-stable? pc)) ports))
-                   (pc (car from-pcs))
-                   (refusals (and instable (port-refusals pc instable))))
+       ;; component LTS is stable when all requires port LTSs are stable
+       ;; and the defer queue is empty.  When the a provides port's
+       ;; modeling-LTS is not stable, the refusals set consists of all
+       ;; its observable events.
+       (let* ((pcs (if (null? traces) from-pcs
+                       (map car traces)))
+              (pc (car pcs)))
+         (and (= (length from-pcs) 1)
+              (null? (.external-q (car from-pcs)))
+              (and (requires-ports-stable? component pc)
+                   (null? (.defer pc)))
+              (let* ((ports (ast:provides-port* component))
+                     (instable (find (negate (port-lts-stable? pc)) ports))
+                     (refusals (and instable (port-refusals pc instable))))
 
-              (define (mark-refusals pc)
-                (clone pc #:status (make <refusals-error>
-                                     #:ast (.behavior component)
-                                     #:message "non-compliance"
-                                     #:refusals refusals)))
+                (define (mark-refusals pc)
+                  (clone pc #:status (make <refusals-error>
+                                       #:ast (.behavior component)
+                                       #:message "non-compliance"
+                                       #:refusals refusals)))
 
-              (and (pair? refusals)
-                   (let ((traces (map (cute rewrite-trace-head mark-refusals <>) traces)))
-                     (report traces
-                             #:eligible '()
-                             #:internal? internal?
-                             #:locations? locations?
-                             #:state? state?
-                             #:trace trace
-                             #:verbose? verbose?))))))))
+                (and (pair? refusals)
+                     (let ((traces (map (cute rewrite-trace-head mark-refusals <>) traces)))
+                       (report traces
+                               #:eligible '()
+                               #:internal? internal?
+                               #:locations? locations?
+                               #:state? state?
+                               #:trace trace
+                               #:verbose? verbose?)))))))))
 
   (let* ((traces (apply append list-of-traces))
          (traces (filter-illegal+implicit-illegal traces))
@@ -1044,7 +1075,8 @@ status."
                       traces))
          (deadlock-check? (and deadlock-check?
                                (not status)))
-         (refusals-check? (and refusals-check?
+         (refusals-check? (and (%compliance-check?)
+                               refusals-check?
                                (not status)
                                (is-a? (%sut) <runtime:component>))))
     (or (and deadlock-check?
