@@ -101,8 +101,7 @@
      ((and (is-a? instance <runtime:component>)
            (or (get-handling pc instance)
                (assq r:port (.blocked pc)))
-           (blocked-port pc instance)
-           (not (blocked-on-boundary? pc)))
+           (blocked-port pc instance))
       (%debug "<blocked-error> collateral ~a ~a\n" (name instance) (trigger->string trigger))
       (let ((message (format #f "port `~a' is collaterally blocked"
                              (.name port))))
@@ -150,6 +149,20 @@
   (%debug "MISSING: ~a ~a\n" ((compose name .instance) pc) (name o))
   '())
 
+(define-method (step (pc <program-counter>) (o <defer>))
+  (let* ((defer (.defer pc)))
+    (if (= (length defer) (%queue-size))
+        (let ((error (make <queue-full-error>
+                       #:ast o
+                       #:instance (.instance pc)
+                       #:message "queue-full")))
+          (list (clone pc #:status error)))
+        (let ((pc (clone pc #:defer (append defer (list pc)))))
+          (list (statement-continuation pc o))))))
+
+(define-method (step (pc <program-counter>) (o <defer-qout>))
+  (list (clone pc #:statement (.statement o))))
+
 (define-method (step (pc <program-counter>) (o <initial-compound>))
   (append
    (next-method pc o)
@@ -384,7 +397,7 @@
                   (ports-eq? (ast:eq? reply-port trigger-port)))
              (and reply-port
                   (or blocking?
-                      (and (or (ast:eq? (.port (rtc-trigger pc)) reply-port)
+                      (and (or (ast:eq? (and=> (rtc-trigger pc) .port) reply-port)
                                (not (and=> (as trigger <q-trigger>) .modeling?)))
                            (or (ast:provides? (rtc-trigger pc))
                                (not (is-a? (ast:type value) <void>)))))
@@ -623,20 +636,31 @@
     (match parent
       (($ <compound>) (statement-continuation pc o))
       (($ <if>) (statement-continuation pc parent))
-      (_ '()))))
+      (($ <defer>)
+       (clone pc #:statement #f))
+      (($ <defer-qout>)
+       (clone pc #:statement #f))
+      (_
+       '()))))
 
 (define-method (statement-continuation (pc <program-counter>) (o <statement>))
-  (let ((parent (.parent o)))
-    (match parent
+  (let ((p (.parent o)))
+    (match p
       (($ <compound>)
-       (let ((next (and=> (member o (ast:statement* parent) ast:eq?) cdr)))
+       (let ((next (and=> (member o (ast:statement* p) ast:eq?) cdr)))
          (if (pair? next) (continuation pc next)
-             (let ((pc (pop-locals pc (filter (is? <variable>) (ast:statement* parent)))))
-               (statement-continuation pc parent)))))
+             (let ((pc (pop-locals pc (filter (is? <variable>) (ast:statement* p)))))
+               (statement-continuation pc p)))))
       (($ <if>)
-       (statement-continuation pc parent))
+       (statement-continuation pc p))
       (($ <function>)
        (pop-pc pc))
+      (($ <defer>)
+       (clone pc #:statement #f))
+      (($ <defer-qout>)
+       (let* ((p (parent p <compound>))
+              (pc (if (not p) pc (pop-locals pc (filter (is? <variable>) (ast:statement* p))))))
+        (clone pc #:statement #f)))
       (_
        '()))))
 
