@@ -61,6 +61,8 @@
             blocked-port
             blocked-ports
             collateral-block
+            defer-event?
+            defer-labels
             dequeue
             dequeue-external
             enqueue
@@ -72,6 +74,7 @@
             get-reply
             get-state
             get-variables
+            graft-locals
             in-event?
             is-status?
             label?
@@ -86,8 +89,10 @@
             pop-locals
             port-event?
             provides-trigger?
+            prune-defer
             push-local
             pc-equal?
+            pc->stack
             pop-deferred
             pop-pc
             push-pc
@@ -173,7 +178,13 @@
          (trail (with-input-from-string trail read-input-file))
          (trail (map ->string trail))
          (loop-index (list-index (cute equal? <> "<loop>") trail))
-         (trail (filter (negate (conjoin string? (cute string-prefix? "<" <>))) trail))
+         (trail (filter (negate (conjoin
+                                 string?
+                                 (conjoin
+                                  (cute string-prefix? "<" <>)
+                                  (negate
+                                   (cute equal? <> "<defer>")))))
+                        trail))
          (loop  (and loop-index (call-with-values
                                     (cute split-at trail loop-index)
                                   (lambda (a b) b))))
@@ -359,7 +370,7 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
        (filter runtime:boundary-port? (%instances)))))
 
 (define-method (labels (pc <program-counter>))
-  (append (labels) (rtc-labels pc) (return-labels pc)))
+  (append (labels) (rtc-labels pc) (return-labels pc) (defer-labels pc)))
 
 (define-method (label? (o <string>))
   (and (member o (labels)) o))
@@ -367,6 +378,10 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
 (define-method (label? (o <boolean>))
   #f)
 
+(define-method (defer-labels (pc <program-counter>))
+  (if (null? (.defer pc)) '()
+      '("<defer>")))
+
 (define-method (return-labels (o <port>))
   (let* ((port (.name o))
          (values (ast:return-values o)))
@@ -376,9 +391,11 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
   (return-labels (.ast o)))
 
 (define-method (rtc-labels (pc <program-counter>))
-  (let* ((released-pc (blocked-on-boundary-collateral-release pc))
+  (let* ((bob-pc (blocked-on-boundary-switch-context pc))
+         (bob-statement (.statement bob-pc))
+         (released-pc (blocked-on-boundary-collateral-release pc))
          (switched-pc (switch-context released-pc)))
-    (if (eq? switched-pc pc) '()
+    (if (or (is-a? bob-statement <trigger-return>) (eq? switched-pc pc)) '()
         (let* ((ports (filter (conjoin runtime:boundary-port? ast:provides?)
                               (%instances)))
                (port-names (map (compose .name .ast) ports)))
@@ -545,12 +562,14 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
 (define-method (push-pc (pc <program-counter>) (instance <runtime:instance>) (statement <statement>))
   (clone pc #:previous pc #:instance instance #:statement statement))
 
+(define-method (pc->stack (pc <program-counter>))
+  (unfold (negate .previous) identity .previous pc))
+
 (define (pc->statements pc)
-  (unfold (negate (conjoin identity .statement)) .statement .previous pc))
+  (map .statement (pc->stack pc)))
 
 (define-method (rtc-pc (pc <program-counter>))
-  (let ((pcs (unfold (negate .previous) identity .previous pc)))
-    (last pcs)))
+  (last (pc->stack)))
 
 (define-method (rtc-block-pc (pc <program-counter>))
   (let loop ((pc pc))
@@ -591,6 +610,7 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
                #:collateral (.collateral pc)
                #:collateral-instance (.collateral-instance pc)
                #:collateral-released (.collateral-released pc)
+               #:defer (.defer pc)
                #:external-q (.external-q pc)
                #:released (.released pc)
                #:state (.state pc)
@@ -700,6 +720,7 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
                     #:id (.id blocked-pc)
                     #:instance (.instance blocked-pc)
                     #:previous (.previous blocked-pc)
+                    #:running-defer? (.running-defer? blocked-pc)
                     #:statement (.statement blocked-pc)
                     #:trigger (.trigger blocked-pc))))))
     (#f
@@ -739,6 +760,7 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
                #:collateral (acons blocked-port pc (.collateral pc))
                #:external-q (.external-q pc)
                #:released (.released pc)
+               #:running-defer? (.running-defer? pc)
                #:state (.state pc)
                #:trail (.trail pc))))
     (%debug "  ~s ~s <collateral-block> ~a [~a] => [~a]\n"
@@ -855,27 +877,56 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
   (set-state pc (clone (get-state pc) #:handling #f)))
 
 (define-method (pop-deferred (pc <program-counter>))
-  (values (set-deferred pc #f) (.deferred (get-state pc))))
+  (let* ((deferred (.deferred (get-state pc)))
+         (queue? (not (q-empty? pc deferred)))
+         (pc (if queue? pc
+                 (set-deferred pc #f))))
+    (values deferred pc)))
 
 (define-method (set-deferred (pc <program-counter>) deferred)
   (set-state pc (clone (get-state pc) #:deferred deferred)))
 
+(define-method (prune-defer (pc <program-counter>))
+  (define (state-equal? pc defer-pc)
+    (define defer-variable-names
+      (map .name (ast:defer-variable* (.statement defer-pc))))
+    (define defer-variable?
+      (match-lambda ((name . expression)
+                     (or (null? defer-variable-names)
+                         (member name defer-variable-names)))))
+    (let* ((instance (.instance defer-pc))
+           (defer-members (get-members defer-pc instance))
+           (defer-members (filter defer-variable? defer-members))
+           (members (get-members pc instance))
+           (members (filter defer-variable? members)))
+      (or (null? defer-variable-names)
+          (ast:equal? members defer-members))))
+  (let* ((defer (.defer pc))
+         (defer (filter (cute state-equal? pc <>) defer)))
+    (clone pc #:defer defer)))
+
 (define-method (flush (pc <program-counter>) instance)
-  (let* ((flush-return (make <flush-return>))
+  (%debug "  ~s ~s ~a\n" (name instance) (and=> (.trigger pc) trigger->string) "<flush>")
+  (let* ((orig-pc pc)
+         (flush-return (make <flush-return>))
          (pc (push-pc pc flush-return))
          (pc (clone pc #:instance instance)))
-    (if (null? (.q (get-state pc))) (let ((pc deferred (pop-deferred pc)))
-                                      (%debug "  flush deferred: ~s\n" deferred)
-                                      (if (not deferred) pc
-                                          (if (get-handling pc deferred) (throw 'handling "already handling event" pc)
-                                              (flush pc deferred))))
-        (let ((pc trigger (dequeue pc)))
-          (let* ((q-out (make <q-out> #:trigger trigger))
-                 (q-out (clone q-out #:location (.location trigger))))
-            (push-pc pc trigger instance q-out))))))
+    (cond
+     ((pair? (.q (get-state pc)))
+      (let ((pc trigger (dequeue pc)))
+        (let* ((q-out (make <q-out> #:trigger trigger))
+               (q-out (clone q-out #:location (.location trigger))))
+          (push-pc pc trigger instance q-out))))
+     (else
+      (let ((deferred pc (pop-deferred pc)))
+        (cond ((or (not deferred)
+                   (get-handling pc deferred))
+               orig-pc)
+              (else
+               (%debug "  flush deferred: ~s\n" deferred)
+               (flush pc deferred))))))))
 
 (define-method (flush (pc <program-counter>))
-  (%debug "  ~s ~s ~a\n" ((compose name .instance) pc) (and=> (.trigger pc) trigger->string) "<flush>")
   (flush pc (.instance pc)))
 
 
@@ -935,6 +986,26 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
 (define-method (set-variables (pc <program-counter>) (o <list>))
   (set-state pc (clone (get-state pc) #:variables o)))
 
+(define-method (get-members (pc <program-counter>) instance)
+  (let* ((state (get-state pc instance))
+         (variables (.variables state))
+         (component (runtime:ast-model instance))
+         (members (ast:member* component)))
+    (take-right variables (length members))))
+
+(define-method (get-locals (pc <program-counter>) instance)
+  (let* ((state (get-state pc instance))
+         (variables (.variables state))
+         (component (runtime:ast-model instance))
+         (members (ast:member* component)))
+    (drop-right variables (length members))))
+
+(define-method (graft-locals (pc <program-counter>) (from <program-counter>))
+  (let* ((instance (.instance from))
+         (members (get-members pc instance))
+         (locals (get-locals from instance)))
+    (set-variables pc (append locals members))))
+
 (define-method (push-local (o <formal>) (e <expression>) (pc <program-counter>))
   (or (and=> (range-error o e) (cut clone pc #:status <>))
       (set-variables pc (acons (.name o) e (get-variables pc)))))
@@ -960,8 +1031,11 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
                (value (.value value)))
            (and (or (< value (.from range))
                     (> value (.to range)))
-                (make <range-error> #:ast o #:variable o #:value value
-                      #:message "range-error"))))))
+                (let ((parent (.parent (.parent o)))
+                      (error (make <range-error>
+                               #:ast o #:variable o #:value value
+                               #:message "range-error")))
+                  (clone error #:parent parent)))))))
 
 (define-method (assign (pc <program-counter>) variable expression)
   (let* ((name (.name variable))
@@ -1153,11 +1227,15 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
             (append
              (let* ((instance (.instance o))
                     (statement (.statement o))
+                    (defer? (or (as statement <defer>)
+                                (as statement <defer-qout>)))
                     (statement (or (as statement <action>)
-                                   (as statement <trigger-return>))))
+                                   (as statement <trigger-return>)
+                                   defer?)))
                (if (not statement) '()
                    `("instance:" ,(runtime:instance->string instance)
-                     "statement:" ,(trigger->string statement)
+                     ,@(if defer? '()
+                           `("statement:" ,(trigger->string statement)))
                      "at:" ,(ast:location->string statement))))
              (let ((triggers (rtc-triggers o)))
                (if (null? triggers) '()
@@ -1209,7 +1287,12 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
                    ,@(map trigger->string (blocked-on-boundary-statements o))))
              (if (null? (.external-q o)) '()
                  (list (external-q->string (.external-q o))))
-             (async-ports o)))
+             (async-ports o)
+             (let ((defer (.defer o)))
+               (if (null? defer) '()
+                   `("defer:"
+                     ,@(map (compose pc->string (cute clone <> #:defer '()))
+                            defer))))))
       "\n"))))
 
 (define-method (pc->string-state-diagram (o <program-counter>))
@@ -1332,6 +1415,10 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
   (or (not (.instance pc))
       (null? (.q (get-state pc)))))
 
+(define-method (q-empty? (pc <program-counter>) instance)
+  (or (not instance)
+      (null? (.q (get-state pc instance)))))
+
 (define-method (rtc-program-counter-equal? (a <program-counter>) (b <program-counter>))
   (and (ast:equal? (.status a) (.status b))
        (ast:eq? (.statement a) (.statement b))
@@ -1383,6 +1470,12 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
        (not (return-trigger? event))
        (pair? (.async pc))))
 
+(define-method (defer-event? (pc <program-counter>) event)
+  (and (string? event)
+       (not (member event (labels)))
+       (not (return-trigger? event))
+       (pair? (.defer pc))))
+
 
 ;;;
 ;;; Initialization
