@@ -1,6 +1,6 @@
 // dzn-runtime -- Dezyne runtime library
 //
-// Copyright © 2014, 2015, 2016, 2017, 2019, 2020, 2021 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
+// Copyright © 2014, 2015, 2016, 2017, 2019, 2020, 2021, 2022 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
 // Copyright © 2015, 2016, 2017, 2019, 2020, 2021, 2022 Rutger van Beusekom <rutger@dezyne.org>
 // Copyright © 2015 Paul Hoogendijk <paul@dezyne.org>
 //
@@ -34,41 +34,43 @@ namespace dzn
   std::ostream debug(nullptr);
   runtime::runtime(){}
 
-  void trace(std::ostream& os, port::meta const& m, const char* e)
+  void trace(std::ostream& os, port::meta const& meta, const char* event_name)
   {
     if (!os.rdbuf())
       return;
-    os << path(m.require.meta, m.require.name) << "." << e << " -> "
-       << path(m.provide.meta, m.provide.name) << "." << e << std::endl;
+    os << path(meta.require.meta, meta.require.name) << "." << event_name << " -> "
+       << path(meta.provide.meta, meta.provide.name) << "." << event_name << std::endl;
   }
 
-  void trace_out(std::ostream& os, port::meta const& m, const char* e)
+  void trace_out(std::ostream& os, port::meta const& meta, const char* event_name)
   {
     if (!os.rdbuf())
       return;
-    os << path(m.require.meta, m.require.name) << "." << e << " <- "
-       << path(m.provide.meta, m.provide.name) << "." << e << std::endl;
+    os << path(meta.require.meta, meta.require.name) << "." << event_name << " <- "
+       << path(meta.provide.meta, meta.provide.name) << "." << event_name << std::endl;
   }
 
-  void trace_qin(std::ostream& os, port::meta const& m, const char* e)
+  void trace_qin(std::ostream& os, port::meta const& meta, const char* event_name)
   {
     if (!os.rdbuf())
       return;
-
-    if (path(m.provide.meta) == "<external>")
-      os << path(m.require.meta, "<q>") << " <- "
-         << path(m.provide.meta, m.provide.name) << "." << e << std::endl;
+    if (path (meta.require.meta, "") == "<external>")
+      trace_out (os, meta, event_name);
     else
-      os <<  path(m.provide.meta, m.provide.name) << ".<q> <- "
-         <<  path(m.require.meta, m.require.name) << "." << e << std::endl;
+      os <<  path (meta.require.meta, "<q>")
+         << " <- "
+         << path (meta.provide.meta, meta.provide.name) << "." << event_name << std::endl;
   }
 
-  void trace_qout(std::ostream& os, port::meta const& m, const char* e)
+  void trace_qout(std::ostream& os, port::meta const& meta, const char* event_name)
   {
     if (!os.rdbuf())
       return;
-    os << path(m.require.meta, m.require.name) << "." << e << " <- "
-       << path(m.require.meta, "<q>") << std::endl;
+    if (path (meta.require.meta, "") == "<external>")
+      return;
+    os << path (meta.require.meta, meta.require.name) << "." << event_name
+       << " <- "
+       << path(meta.require.meta, "<q>") << std::endl;
   }
 
   bool runtime::external(void* scope) {
@@ -80,6 +82,11 @@ namespace dzn
     return states[scope].handling;
   }
 
+  size_t& runtime::blocked(void* scope)
+  {
+    return states[scope].blocked;
+  }
+
   void*& runtime::deferred(void* scope)
   {
     return states[scope].deferred;
@@ -123,17 +130,17 @@ namespace dzn
     }
   }
 
-  void runtime::defer(void* src, void* tgt, const std::function<void()>& event, size_t coroutine_id)
+  void runtime::enqueue(void* source, void* target, const std::function<void()>& event, size_t coroutine_id)
   {
-    if(!(src && performs_flush(src)) && !handling(tgt))
+    if(!(source && performs_flush(source)) && !handling(target))
     {
-      handle(tgt, event, coroutine_id);
-      flush(tgt, coroutine_id);
+      handle(target, event, coroutine_id);
+      flush(target, coroutine_id);
     }
     else
     {
-      deferred(src) = tgt;
-      queue(tgt).push(event);
+      deferred(source) = target;
+      queue(target).push(event);
     }
   }
 }
