@@ -42,6 +42,34 @@ namespace dzn
   public class pump : IDisposable
   {
     int id = 0;
+    public static void defer(Locator l, Object c, Func<bool> p, Action<int> f)
+    {
+      l.get<pump>().defer(p, f);
+    }
+    public static void prune_deferred(Locator l)
+    {
+      pump p = l.try_get<dzn.pump>();
+      if (p != null)
+        p.prune_deferred();
+    }
+    public static int coroutine_id(dzn.Locator l)
+    {
+      pump p = l.try_get<dzn.pump>();
+      return p == null ? 1 : p.coroutine_id();
+    }
+    public static void port_block(Locator l, Object c, Object p)
+    {
+      l.get<pump>().block(l.get<Runtime>(), c, p);
+    }
+    public static void port_release(Locator l, Object p)
+    {
+      var pump = l.get<pump>();
+      pump.release(l.get<dzn.Runtime>(),p);
+    }
+    public static void collateral_block(Object c, dzn.Locator l)
+    {
+      l.get<dzn.pump>().collateral_block(c, l.get<dzn.Runtime>());
+    }
     public static bool port_blocked_p(Locator loc, Object p)
     {
       dzn.pump pump = loc.try_get<pump>();
@@ -49,17 +77,6 @@ namespace dzn
           return pump.blocked_p(p);
       return false;
     }
-    public static void port_block(Locator l, Object c, Object p)
-    {
-      l.get<pump>().block(l.get<Runtime>(), c, p);
-    }
-
-    public static void port_release(Locator l, Object p)
-    {
-      var pump = l.get<pump>();
-      pump.release(l.get<dzn.Runtime>(),p);
-    }
-
     public static coroutine find_self(list<coroutine> coroutines)
     {
       var count = coroutines.FindAll(c => c.port == null && !c.finished).Count;
@@ -100,9 +117,9 @@ namespace dzn
       {
         Deadline d = (Deadline)o;
         if(rank == d.rank && t == d.t && id == d.id) return 0;
-        if(rank < d.rank ||
-           rank == d.rank && t < d.t ||
-           rank == d.rank && t == d.t && id < d.id) return -1;
+        if(rank < d.rank
+           || rank == d.rank && t < d.t
+           || rank == d.rank && t == d.t && id < d.id) return -1;
         return 1;
       }
     };
@@ -114,13 +131,16 @@ namespace dzn
     public list<coroutine> collateral_blocked = new list<coroutine>();
     public List<Object> skip_block = new List<Object>();
     public Queue<Action> queue = new Queue<Action>();
+    public Queue<Tuple<Func<bool>,Action<int>>> deferred = new Queue<Tuple<Func<bool>,Action<int>>>();
     public bool running;
+    public bool paused;
     public List<Object> unblocked = new List<Object>();
     public Thread task;
 
     public pump()
     {
       this.running = true;
+      this.paused = false;
       this.task = new Thread(this.run);
       this.task.Start();
     }
@@ -158,6 +178,28 @@ namespace dzn
           }
         });
     }
+    public void pause ()
+    {
+      Debug.WriteLine ("pump.pause");
+      context.lck (this, () => {
+        this.paused = true;
+      });
+    }
+    public void resume ()
+    {
+      Debug.WriteLine ("pump.resume");
+      context.lck (this, () => {
+        this.paused = false;
+        Monitor.Pulse (this);
+      });
+    }
+    public void flush ()
+    {
+      Debug.WriteLine ("pump.flush");
+      resume ();
+      System.Threading.Thread.Sleep (100);
+      pause ();
+    }
     public void run()
     {
       try {
@@ -166,8 +208,8 @@ namespace dzn
               if(this.queue.Count == 0) {
                 Monitor.Pulse(this);
               }
-              if(this.timers.Count == 0) {
-                while(this.queue.Count == 0 && running) {
+              if (this.timers.Count == 0 && this.deferred.Count == 0) {
+                while (this.queue.Count == 0 && this.deferred.Count == 0 && running) {
                   Monitor.Wait(this);
                 }
               } else {
@@ -175,7 +217,7 @@ namespace dzn
                 t.MoveNext();
                 bool timedout = false;
                 TimeSpan wait = t.Current.Key.t - DateTime.Now;
-                while(!timedout && this.queue.Count == 0 && running && wait.Ticks > 0) {
+                while(!timedout && this.queue.Count == 0 && this.deferred.Count == 0 && running && wait.Ticks > 0) {
                   timedout = !Monitor.Wait(this, wait);
                 }
               }
@@ -186,6 +228,15 @@ namespace dzn
                 f();
               }
 
+              if (this.queue.Count == 0 && this.deferred.Count != 0 && this.deferred.Peek().Item1())
+              {
+                var t = this.deferred.Peek();
+                this.deferred.Dequeue();
+                if (Monitor.IsEntered(this)) Monitor.Exit(this);
+                t.Item2(id);
+                Monitor.Enter(this);
+              }
+
               {
                 IEnumerator<KeyValuePair<Deadline, Action>> t = timers.OrderBy(k => k.Key).GetEnumerator();
                 while(timers.Count != 0 && t.MoveNext() && t.Current.Key.expired()) {
@@ -205,6 +256,7 @@ namespace dzn
         context.lck(this, () => {
             while(this.running || this.queue.Count!=0 || this.collateral_blocked.Count!=0)
             {
+              while (this.paused) Monitor.Wait(this);
               Monitor.Exit(this);
               Debug.Assert(this.coroutines.Count!=0);
               this.coroutines.Last().call(zero);
@@ -259,10 +311,6 @@ namespace dzn
         context();
       }
     }
-    public static void collateral_block(Object c, dzn.Locator l)
-    {
-      l.get<dzn.pump>().collateral_block(c, l.get<dzn.Runtime>());
-    }
     public void collateral_block(Object c, Runtime rt)
     {
       coroutine self = find_self(this.coroutines);
@@ -273,34 +321,21 @@ namespace dzn
       this.coroutines.Remove(self);
 
       self.component = c;
-      Debug.Assert(self.port == null);
-      foreach(var id in rt.blocked_port_component_stack.Keys)
-        if(rt.blocked_port_component_stack[id].Contains(c))
-        {
-          int i = this.coroutines.FindIndex(o => o.id == id);
-          self.port = this.coroutines[i].port;
-        }
-      Debug.Assert(self.port != null, "no port found associated to component " + c.GetHashCode());
 
-      var v = rt.blocked_port_component_stack.ContainsKey(self.id)
-            ? rt.blocked_port_component_stack[self.id]
-            : new Stack<Object>();
-      foreach(var o in v.ToArray().Reverse())
-        rt.component_stack.Push(o);
-      v.Clear();
-      Debug.Assert(!rt.blocked_port_component_stack.ContainsKey(self.id)
-                   || rt.blocked_port_component_stack[self.id].Count == 0);
-      rt.blocked_port_component_stack[self.id] = rt.component_stack;
-      rt.component_stack = v;
+      int coroutine_id = rt.states[c].handling | rt.states[c].blocked;
+      var it = coroutines.Find(coroutine => coroutine.id == coroutine_id);
+
+      if(it == null || it.port == null)
+        throw new RuntimeException("blocking port not found");
+
+      self.port = it.port;
+
+      Debug.WriteLine("[" + self.id + "] collateral block on "
+                      + self.port.GetHashCode());
 
       create_context();
       self.yield_to(this.coroutines.Last());
       Debug.WriteLine("[" + self.id + "] collateral_unblock");
-
-      v = rt.blocked_port_component_stack[self.id];
-      foreach (var o in v.Reverse ())
-        rt.component_stack.Push (o);
-      rt.blocked_port_component_stack[self.id].Clear ();
     }
     public void collateral_release(coroutine self)
     {
@@ -340,18 +375,13 @@ namespace dzn
     }
     public void block(Runtime rt, Object c, Object p)
     {
+      coroutine self = find_self(this.coroutines);
+      rt.states[c].blocked = self.id;
       rt.states[c].handling = 0;
       rt.flush(c, this.coroutine_id());
       if (this.skip_block.Remove(p))
         return;
 
-      coroutine self = find_self(this.coroutines);
-      Debug.Assert(!rt.blocked_port_component_stack.ContainsKey(self.id)
-                   || rt.blocked_port_component_stack[self.id].Count == 0);
-
-      rt.blocked_port_component_stack[self.id] = rt.component_stack;
-      rt.component_stack = new Stack<Object>();
-
       self.port = p;
       Debug.WriteLine("[" + self.id + "] block on " + p.GetHashCode());
 
@@ -376,6 +406,8 @@ namespace dzn
       Debug.WriteLine("");
 
       remove_finished_coroutines(this.coroutines);
+      this.skip_block.Remove(p);
+      rt.states[c].blocked = 0;
     }
     bool collateral_release_skip_block(Runtime rt, Object c)
     {
@@ -384,22 +416,18 @@ namespace dzn
       int it = 0;
       while(it < this.collateral_blocked.Count())
       {
-        coroutine zelf = this.collateral_blocked[it++];
-        if (this.unblocked.FindIndex(i => i == zelf.port) != -1
-            && zelf.component == c)
+        coroutine self = this.collateral_blocked[it++];
+        if (this.unblocked.FindIndex(i => i == self.port) != -1
+            && self.component == c)
         {
-          Debug.WriteLine("[" + zelf.id + "]" + "relay skip "
-                          + zelf.port.GetHashCode());
-          //swap
-          var v = rt.blocked_port_component_stack[zelf.id];
-          rt.blocked_port_component_stack[zelf.id] = rt.component_stack;
-          rt.component_stack = v;
+          Debug.WriteLine("[" + self.id + "]" + "relay skip "
+                          + self.port.GetHashCode());
           have_collateral = true;
-          zelf.component = null;
-          zelf.port = null;
+          self.component = null;
+          self.port = null;
           //splice
-          this.coroutines.Add(zelf);
-          this.collateral_blocked.Remove(zelf);
+          this.coroutines.Add(self);
+          this.collateral_blocked.Remove(self);
         }
       }
       collateral_blocked.Reverse();
@@ -423,28 +451,17 @@ namespace dzn
       Debug.WriteLine("[" + blocked.id + "] unblock");
 
       this.switch_context.Add (() => {
-        var zelf = find_self(this.coroutines);
+        self = find_self(this.coroutines);
         Debug.WriteLine("setting unblocked to port " + blocked.port.GetHashCode());
         this.unblocked.Add(blocked.port);
         blocked.component = null;
         blocked.port = null;
 
-        Debug.WriteLine("[" + zelf.id + "] switch from");
+        Debug.WriteLine("[" + self.id + "] switch from");
         Debug.WriteLine("[" + blocked.id + "] to");
 
-        Debug.Assert(rt.component_stack.Count == 0);
-
-        if(p == null) Console.Error.WriteLine("null port");
-        if(!rt.blocked_port_component_stack.ContainsKey(blocked.id))
-          Console.Error.WriteLine("id " + blocked.id + " not found");
-
-        //swap
-        var v = rt.blocked_port_component_stack[blocked.id];
-        rt.blocked_port_component_stack[blocked.id] = rt.component_stack;
-        rt.component_stack = v;
-
-        zelf.finished = true;
-        zelf.yield_to(blocked);
+        self.finished = true;
+        self.yield_to(blocked);
         Debug.Assert(false, "we must never return here!!!");
         });
     }
@@ -456,6 +473,14 @@ namespace dzn
           Monitor.Pulse(this);
         });
     }
+    public void defer(Func<bool> p, Action<int> e)
+    {
+      this.deferred.Enqueue(new Tuple<Func<bool>,Action<int>> (p, e));
+    }
+    public void prune_deferred()
+    {
+      this.deferred = new Queue<Tuple<Func<bool>,Action<int>>>(this.deferred.Where (t => t.Item1()));
+    }
     public void handle(int id, int ms, Action e)
     {
       handle(id, ms, e, int.MaxValue);
